
import os
import warnings
from pandas import Series,DataFrame
import pandas as pd
import xml.etree.ElementTree as ET

class BroGmwXml:
    """Read well construction data from XML"""

    NS = {
        'gco' : "http://www.isotc211.org/2005/gco",
        'dsgmw' : "https://schema.broservices.nl/xsd/dsgmw/1.1",
        'ns12' : "http://www.broservices.nl/xsd/gmwcommon/1.1",
        'ns11' : "http://www.broservices.nl/xsd/dsgmw/1.1",
        'brocom' : "https://schema.broservices.nl/xsd/brocommon/3.0",
        'gts' : "http://www.isotc211.org/2005/gts",
        'gml' : "http://www.opengis.net/gml/3.2",
        'ns10' : "http://www.broservices.nl/xsd/brocommon/3.0",
        'gmwcom' : "https://schema.broservices.nl/xsd/gmwcommon/1.1",
        'isgmw' : "https://schema.broservices.nl/xsd/isgmw/1.1",
        'xlink' : "http://www.w3.org/1999/xlink",
        'gmd' : "http://www.isotc211.org/2005/gmd",
        }

    WELLPROPTAGS = [
        ('broId','ns10:broId'),   
        ('deliveryAccountableParty','ns10:deliveryAccountableParty'),
        ('qualityRegime','ns10:qualityRegime'),
        ('withPrehistory','ns11:withPrehistory'),
        ('deliveryContext','ns11:deliveryContext'),
        ('constructionStandard','ns11:constructionStandard'),
        ('initialFunction','ns11:initialFunction'),
        ('removed','ns11:removed'),
        ('numberOfMonitoringTubes','ns11:numberOfMonitoringTubes'),
        ('groundLevelStable','ns11:groundLevelStable'),
        ('wellStability','ns11:wellStability'),
        ('nitgCode','ns11:nitgCode'),
        ('wellCode','ns11:wellCode'),
        ('owner','ns11:owner'),
        ('wellHeadProtector','ns11:wellHeadProtector'),
        ('location','ns12:location/gml:pos'),
        ('horizontalPositioningMethod','ns12:horizontalPositioningMethod'),
        ('localVerticalReferencePoint','ns12:localVerticalReferencePoint'),
        ('VerticalOffset','ns12:offset'),
        ('verticalDatum','ns12:verticalDatum'),
        ('groundLevelPosition','ns12:groundLevelPosition'),
        ('groundLevelPositioningMethod','ns12:groundLevelPositioningMethod'),
        ('standardizedLocation','ns11:standardizedLocation/ns10:location/gml:pos'),
        ('coordinateTransformation','ns10:coordinateTransformation'),
        ('objectRegistrationTime','ns10:objectRegistrationTime'),
        ('registrationStatus','ns10:registrationStatus'),
        ('latestAdditionTime','ns10:latestAdditionTime'),
        ('wellConstructionDate','ns11:wellConstructionDate/ns10:date'),
        ]

    TUBEPROPTAGS = [
        'ns11:tubeNumber','ns11:tubeType','ns11:artesianWellCapPresent',
        'ns11:sedimentSumpPresent','ns11:numberOfGeoOhmCables','ns11:tubeTopDiameter',
        'ns11:variableDiameter','ns11:tubeStatus','ns11:tubeTopPosition',
        'ns11:tubeTopPositioningMethod','ns11:tubePartInserted','ns11:tubeInUse',
        'ns12:tubePackingMaterial','ns12:tubeMaterial','ns12:glue',
        'ns11:screenLength','ns11:sockMaterial','ns11:screenTopPosition',
        'ns12:plainTubePartLength','ns11:screenBottomPosition',
        ]


    def __init__(self,tree):
        """
        Parameters
        ----------
        tree : ElementTree tree object
            Valid XML tree with groundwater level measurement data.

        Example
        -------
        gld = BroGld(<elementtree>)
            
        """       
        self._tree = tree
        self._root = self._tree.getroot()
        
    def __repr__(self):
        return self.wellprops['wellCode']

    @classmethod
    def from_xml(cls,xmlpath):
        """Read BRO GMW object from XML file.
        
        Parameters
        ----------
        xmlpath : str
            Valid filepath to XML file with well construction data.

        Returns
        -------
        BroGmwXml object

        Example
        -------
        gmw = BroGmwXml.from_xml(<valid xml filepath>)
        """

        if not os.path.isfile(xmlpath):
            raise ValueError(f'Invalid filepath: "{xmlpath}".')
        tree = ET.parse(xmlpath)
        return cls(tree)

    @property
    def wellprops(self):
        """Read well properties from XML tree"""
        wellprops = {}
        for key,tag in self.WELLPROPTAGS:
            node = self._root.find(f'.//{tag}',self.NS)
            if node is not None:
                wellprops[key]=node.text
            else:
                wellprops[key]=''
                #warnings.warn(f'{key} not found in XML {self._xmlpath}')

        # split coordinates
        wellprops = Series(wellprops)
        wellprops['xcr'] = wellprops['location'].split()[0]
        wellprops['ycr'] = wellprops['location'].split()[1]        
        wellprops['lon'] = wellprops['standardizedLocation'].split()[0]
        wellprops['lat'] = wellprops['standardizedLocation'].split()[1]
        wellprops = wellprops.drop(labels=['location','standardizedLocation',])
        
        return wellprops

    @property
    def tubeprops(self):
        """Read tube properties from XML tree"""
        tubeprops = []
        for node in self._root.iterfind(".//ns11:monitoringTube", self.NS):
            props = {}
            for tag in self.TUBEPROPTAGS:
                property = node.find(f'.//{tag}',self.NS).text
                key = tag.split(':')[1]     
                props[key] = property
            tubeprops.append(props)
        tubeprops = DataFrame(tubeprops)
        tubeprops = tubeprops.set_index('tubeNumber',drop=True)
        return tubeprops

    @property
    def events(self):
        """Read tube changes from XML tree"""
        events = []
        for node in self._root.iterfind('.//ns11:intermediateEvent',self.NS):
            event = {
                'intermediateEvent':node.find(f'.//ns11:eventName',self.NS).text,
                'eventDate':node.find(f'.//ns10:date',self.NS).text,
                }
            events.append(event)
        return DataFrame(events)






import pytest
from pandas import Series, DataFrame
import pandas as pd
import matplotlib
from acequia import Quantiles
from acequia import GwSeries

@pytest.fixture
def gw():
    fpath = r'.\data\dinogws\B21A0138001_1.csv'
    return GwSeries.from_dinogws(fpath)


def test_init(gw):

    # input gwseries object
    qt = Quantiles(gw)
    assert qt.__class__.__name__=='Quantiles'

    # input pandas Series object
    sr = gw.heads(ref='datum')
    qt = Quantiles(sr)
    assert qt.__class__.__name__=='Quantiles'

    # input pandas DataFrame object
    df = DataFrame(gw.heads(ref='datum'))
    qt = Quantiles(sr)
    assert qt.__class__.__name__=='Quantiles'
    

@pytest.fixture
def qt():
    fpath = r'.\data\dinogws\B21A0138001_1.csv'
    gw = GwSeries.from_dinogws(fpath)
    return Quantiles(gw)

@pytest.fixture
def qt2():
    fpath = r'.\data\dinogws\B01D0143001_1.csv'
    gw = GwSeries.from_dinogws(fpath)
    return Quantiles(gw)


def test_quantiles(qt):

    # unit is days, step is valid
    quantiles = qt.get_quantiles(unit='days',step=30)
    assert isinstance(quantiles,DataFrame)
    assert not quantiles.empty

    # unit is days, step is invalid, default is used
    quantiles = qt.get_quantiles(unit='days',step=400)
    assert isinstance(quantiles,DataFrame)
    assert not quantiles.empty
    
    # unit is quantiles, step is valid
    quantiles = qt.get_quantiles(unit='quantiles',step=0.3)
    assert isinstance(quantiles,DataFrame)
    assert not quantiles.empty

    # unit is quantiles, step is invalid, default is used
    quantiles = qt.get_quantiles(unit='quantiles',step=3.0)
    assert isinstance(quantiles,DataFrame)
    assert not quantiles.empty

    # unit is invalid, exception is raised
    with pytest.raises(Exception):
        quantiles = qt.get_quantiles(unit='invalid_input')

def test_headsref(qt):
    assert isinstance(qt.headsref,str)

def test_summary(qt):

    summary = qt.get_summary()
    assert isinstance(summary,DataFrame)
    assert not summary.empty

def test_get_inundation(qt2):

    sr = qt2.get_inundation()
    assert isinstance(sr,Series)
    assert not sr.empty

def test_get_lowest(qt2):

    sr = qt2.get_lowest()
    assert isinstance(sr,Series)
    assert not sr.empty

def test_repr(qt):
    assert isinstance(str(qt),str)

def test_plot(qt):

    # plot with defaults
    myplot = qt.plot()
    assert isinstance(myplot,matplotlib.axes.Subplot)
    
    # test parameter coloryears
    myplot = qt.plot(coloryears=[2002,2007])
    assert isinstance(myplot,matplotlib.axes.Subplot)

    # test boundyears
    myplot = qt.plot(boundyears=[2004,2005])
    assert isinstance(myplot,matplotlib.axes.Subplot)
    
    # test figpath
    myplot = qt.plot(figpath=r'.\output\fig\duurlijnen.jpg')
    assert isinstance(myplot,matplotlib.axes.Subplot)





import pytest
##import collections
##import numpy as np
from pandas import DataFrame
import pandas as pd

from acequia import Dawaco, GwSeries
import acequia as aq

srcpath = r'.\data\dawaco\small_dataset.xlsx'

@pytest.fixture
def daw():
    # test_from_excel()
    daw = Dawaco.from_excel(srcpath)
    assert isinstance(daw, Dawaco)
    assert isinstance(daw.rawdata, DataFrame)
    return daw

def test_len(daw):
    assert len(daw)!=0

def test_repr(daw):
    assert isinstance(repr(daw),str)

def test_filters(daw):
    filters = daw.filters
    assert isinstance(filters, list)
    assert len(filters)!=0

def test_get_gwseries(daw):
    loc, fil = daw.filters[0]
    gw = daw.get_gwseries(loc,fil)
    assert isinstance(gw, GwSeries)
    assert len(gw)!=0

def test_iteritems(daw):
    for gw in daw.iteritems():
        assert isinstance(gw, GwSeries)





import warnings
import numpy as np
import matplotlib.pyplot as plt
from pandas import Series, DataFrame
import pandas as pd


class MeteoDrought:
    """Calculate statistical measures of meteorological drought from
    time series

    Parameters
    ----------
    prc : pd.Series
        time series with precipitation data

    evp : pd.Series
        time series with evaporation data

    Examples
    --------

    dro = aq.MeteoDrought(prc,evp)

    drday = dro.daydrought()

    drsum = dro.summersum()

    drcum = dro.summercum()

    """

    YEARSTART = 'YEAR-01-01'
    YEAREND = 'YEAR-12-31'
    SUMMERSTART = 'YEAR-04-01'
    SUMMEREND = 'YEAR-09-30'
    SUMMERDAYS = 183


    def __init__(self,prc=None,evp=None,stn=None):

        self._prc = prc
        self._evp = evp
        self._rch = self.recharge()
        self._rchsmr = self.summer_recharge()

        if stn is None:
            stn = 'unknown'
        self._stn = stn


    def __repr__(self):
        return (f'{self.__class__.__name__} (n={len(self._rchsmr)})')


    def recharge(self):
        """Return time series of recharge for all available days

        Calculated as difference between precipitation and evaporation.
        Leading and trailing NaN values are removed."""

        self._rch = self._prc - self._evp

        # remove leading and trailing NaNs
        first = self._rch.first_valid_index()
        last = self._rch.sort_index(ascending=False).first_valid_index()
        self._rch = self._rch[first:last]

        # remove first year if first date is after april 1st
        if not self._rch.index[0].month<4:
            firstyear = self._rch.index[0].year+1
            firstdate = pd.to_datetime(
                self.YEARSTART.replace('YEAR',str(firstyear)))
            self._rch = self._rch[firstdate:]

        # remove last year is last date is before september 30th
        if not self._rch.index[-1].month>9:
            lastyear = self._rch.index[-1].year-1
            lastdate = pd.to_datetime(
                self.YEAREND.replace('YEAR',str(lastyear)))
            self._rch = self._rch[:lastdate]

        return self._rch


    def summer_recharge(self):
        """Return table with array of daily recharges for each summer"""

        # empty table
        years = list(set(self._rch.index.year))
        days = np.arange(0,183)
        self._rchsmr = Series(index=years,dtype=object)
        self._rchsmr.index.name = 'year'

        # daily rechsrge for all years
        for year,rch in self._rch.groupby(by=self._rch.index.year):

            firstdate = self.SUMMERSTART.replace('YEAR',str(year))
            lastdate = self.SUMMEREND.replace('YEAR',str(year))
            rchsmr = self._rch[firstdate:lastdate].values

            self._rchsmr[year] = rchsmr

        return self._rchsmr


    def _cumulative_drought(self,rchsmr):
        """Return daily values of cumulative drought for one summer

        Parameters
        ----------
        rchsmr : np.array
            daily values of recharge between april 1st and september 30th

        Returns
        -------
        numpy.array with daily cumulative drought 

        """

        daydr = -1*rchsmr
        cumdr = np.zeros(len(daydr))

        for i,val in enumerate(daydr):

            if i==0:

                if np.isnan(daydr[i]):
                    cumdr[i] = daydr[i]

                elif daydr[i] > 0:
                    cumdr[i] = daydr[i]

                else:
                    cumdr[i] = 0

            else:
                cumdr[i] = cumdr[i-1] + daydr[i]
                if cumdr[i]<0:
                    cumdr[i]=0

        return cumdr


    def daydrought(self):
        """Return cumulative drought on daily basis for all years

        Returns
        -------
        pd.DataFrame

        Notes
        -----
        Cumulative meteorological drought is calculated between april 1st
        and september 30th.

        """

        years = list(set(self._rch.index.year))
        days = np.arange(1,self.SUMMERDAYS+1)
        self._daydrought = DataFrame(columns=years,index=days)
        self._daydrought.index.name = 'daynr'
        for year,rch in self._rchsmr.iteritems():
            self._daydrought[year] = self._cumulative_drought(rch)

        return self._daydrought


    def summercum(self):
        """Return maximum cumulative drought for each year"""
        return self.daydrought().max(axis=0)


    def summersum(self):
        """Return sum of drought for all years"""

        self._summersum = Series(index=self._rchsmr.index,dtype=float)

        for year,rch in self._rchsmr.iteritems():
            self._summersum[year] = np.sum(-1*rch)

        return self._summersum




"""

GwList is a class that holds a list of GwSeries objects and 
provides methodes for iterating
=======
GwList
======
Object for maintaining a groundwater series iterator


""" 

import os
import os.path
import warnings
from pandas import Series, DataFrame
import pandas as pd
import time
from datetime import datetime

from . import gwseries as gwseriesmod
from .read.hydromonitor import HydroMonitor
from .read.waterweb import WaterWeb
from .read.filedirtools import listdir

newline = '\n'

def headsfiles(srcdir=None,srctype=None,loclist=None):
    """Return list of sourcefiles in directory

    Parameters
    ----------
    srcdir : str
        directory with sourcefiles

    srctype : {'dinocsv','json'}, optional
        sourcefiletype

    loclist : list, optional
        list of strings with location names

    Return
    ------
    pd.DataFrame with series sourcefilelist

    """
    gws = GwList(srcdir=srcdir,srctype=srctype,loclist=loclist)
    return gws.filetable()


class GwList():
    """Container for a sequence of GwSeries objects
    
    Examples  
    --------  
    Create GwList object and load multiple sourcefiles:
    >>>gwl = GwList(srcdir=<sourcedir>,srctype='dinocsv',loclist=<mylist>)  
    >>>gwl = GwList(srcdir=<sourcedir>,srctype='json',loclist=<mylist>)    
    >>>gwl = GwList(srcfile=<filepath>,srctype=<'json' or ' dinocsv'>)      
    >>>gwl = GwList(srfile=<hydromonitor csv export>,srctype='hymon')  
    >>>gwl = GwList(srfile=<waterweb csv export>,srctype='waterweb')
  
    Return table with location properties:
    >>>locp = gwl.locprops()

    Return list of soourcefiles in <srcdir> of type ' dinocsv':
    >>>gwl.filelist()

    Notes
    -----
    When only srcdir is given, result will be a list of
    GwSeries objects for all sourcefiles in srcdir.

    When both srcdir and srcfile are given, all files from srcdir will
    be selected and srcfile will be ignored.

    When loclist is given, names in this list will be used for 
    selecting files in srcdir. All series that belong to a location
    will be selected as seperate series.
       
    """


    _valid_srctype = ['dinocsv','json','hymon','waterweb']


    def __init__(self,srcdir=None,srctype='dinocsv',loclist=None,
        srcfile=None):
        """
        Parameters
        ----------
        srcdir : str
            directory with groundwater head sourcefiles
        srctype : {'dinocsv','json','hymon', 'waterweb'}, optional
            sourcefiletype
        loclist : list, optional
            list of location names
        srcfile : str, optional
            path to file with paths to sourcefiles
            (if srcfile is given, srcdir is ignored)

        """
        message = ('Class GwList will be deprecated in a future version.'
            'Use GwCollection class instead.')
        warnings.warn(message, DeprecationWarning, stacklevel=2)

        self._srcdir = srcdir
        self._srctype = srctype
        self._loclist = loclist
        self._srcfile = srcfile
        self._itercount = 0


        if (self._srcdir is None) and (self._srcfile is None):
            raise ValueError((
                'Missing parameter value. '
                f'At least one of parameters srcdir or '
                f'srclist must be given.'))

        if self._srctype not in self._valid_srctype:
            raise ValueError((
                'srctype value not recognised.\n'
                f'"{self._srctype}" is not a valid '
                f'sourcefile type. Valid sourcefiletypes are '
                f'{self._valid_srctype}'))

        if (self._srcdir is not None) and (self._srcfile is not None):
            self._srcfile = None # given value for srcfile is ignored!
            warnings.warn((
                'Ambigious combination of parameter values:\n'
                f'srcdir is {self._srcdir} (not None) and srcfile is '
                f'{self._srcfile} (not None). Given value for srcfile '
                f'will be ignored.'))

        if self._srcdir is not None:

            if not os.path.isdir(self._srcdir):
                raise ValueError(f'Directory {self._srcdir} does not exist')

            self._flist = self.filetable() #_sourcefiles()


        if (self._srcfile is not None) and (
            self._srctype in ['dinocsv','json']):

            if not os.path.exists(self._srcfile):
                raise ValueError(f'Filepath {self._srcfile} does not exist.')

            self._flist = self.filetable()

        if (self._srcfile is not None) and (self._srctype=='hymon'):
            self.hm = HydroMonitor(fpath=srcfile)

        if (self._srcfile is not None) and (self._srctype=='waterweb'):
            self._wwn = WaterWeb.from_csv(srcfile,network=None)

    def __repr__(self):
        mylen = len(self)
        return (f'{self.__class__.__name__}({mylen})')

    def filetable(self):
        """ Return list of sourcefile names """

        if (self._srcdir is not None) and (self._srctype 
                in ['dinocsv','json']):

            return self._sourcefiles()

        if (self._srcfile is not None) and (self._srctype in ['dinocsv','json']):

            ftime = datetime.fromtimestamp(os.path.getmtime(self._srcfile))
            fileage = datetime.now()-ftime
            if fileage.days > 1:
                msg = f'Age of {self._srcfile} is {fileage.days} days.' 
                warnings.warn(msg)

            #TODO: check if flist contains valid sourcefilenames

            ## flist must be a list of sourcefilesnames
            flist = pd.read_csv(self._srcfile,index_col=False)

            return flist

        if (self._srcfile is not None) and (self._srctype=='hymon'):
            warnings.warn((
                f'function filelist() not supported for sourcetype'
                f' "hymon" '))
            ##logger.warning(msg)
            return None

        warnings.warn((f'Unexpected combination of given parameters. '
            f'No list of GwSeries objects is returned.'))
        ##logger.warning(msg)
        return None


    def __iter__(self):
        """ return iterator """
        return self


    def __next__(self):
        """ return next gwseries object in list """
        
        if self._itercount >= self.__len__():
            self._itercount = 0 # reset iterator
            raise StopIteration

        if self._srctype == 'dinocsv':
            idx = self._flist.index[self._itercount]
            filename = self._flist.at[self._itercount,'path']
            self.gw = gwseriesmod.GwSeries.from_dinogws(filename)
 
        if self._srctype == 'json':
            idx = self._flist.index[self._itercount]
            filename = self._flist.at[idx,'path']
            self.gw = gwseriesmod.GwSeries.from_json(filename)

        if self._srctype == 'hymon':
            self.gw = next(self.hm)

        if self._srctype == 'waterweb': 
            srname = self._wwn.srnames()[self._itercount]
            self.gw = self._wwn.gwseries(srname)

        self._itercount += 1
        return self.gw

   
    def __len__(self):

        if self._srctype in ['dinocsv','json']:
            return len(self._flist)

        if self._srctype in ['hymon']:
            return len(self.hm)

        if self._srctype=='waterweb':
            return len(self._wwn)


    def is_callable(self):
        """Return True if object is waiting for a call"""

        if self._itercount==0:
            is_callable = True
        else:
            is_callable = False
        return is_callable


    def _sourcefiles(self):
        """ return list of sourcefiles in directory dir"""

        if self._srctype=='dinocsv':

            pathlist = listdir(self._srcdir, filetype='csv')
            filelist = [os.path.split(path)[-1] for path in pathlist if path.split('_')[-1].endswith('1.csv')]
            dnfiles = pd.DataFrame({"file":filelist})

            dnfiles["loc"] = dnfiles["file"].apply(lambda x:x[0:8])
            dnfiles["fil"] = dnfiles["file"].apply(
                lambda x:x[8:11].lstrip("0"))
            dnfiles["kaartblad"] = dnfiles["loc"].apply(lambda x:x[1:4])
            dnfiles["series"]= dnfiles["loc"]+"_"+dnfiles["fil"]
            dnfiles["path"]= dnfiles["file"].apply(lambda x:self._srcdir+x)

            if self._loclist is not None:
                mask = dnfiles['loc'].isin(self._loclist)
                dnfiles = dnfiles[mask].reset_index(drop=True)

            return dnfiles

        if self._srctype=='json':

            files = listdir(self._srcdir, filetype='json')

            jsf = pd.DataFrame({"file":files})
            jsf["series"]= jsf["file"].apply(lambda x:x.split('.')[0])
            jsf["loc"] = jsf["series"].apply(lambda x:x.split('_')[0])
            jsf["fil"] = jsf["series"].apply(
                         lambda x:x.split('_')[-1].lstrip("0"))

            jsf["path"]= jsf['file'].apply(lambda x:os.path.join(
                                                    self._srcdir,x))

            if self._loclist is not None:
                mask = jsf['loc'].isin(self._loclist)
                jsf = jsf[mask].reset_index(drop=True)

            return jsf


    def gwseries(self,srname):
        """ Return single named GwSeries object from list

        Parameters
        ----------
        srname : str
            series name of requested GwSeries object

        Returns
        -------
        GwSeries object
        """

        if self._srctype in ['dinocsv','json']:
            row = self._flist[self._flist['series']==srname]
            indexval = row.index.values[0]
            filepath = self._flist.loc[indexval,'path']

        if self._srctype=='json':
            gw = gwseriesmod.GwSeries.from_json(filepath)

        if self._srctype=='dinocsv':
            gw = gwseriesmod.GwSeries.from_dinogws(filepath)

        if self._srctype=='hymon':
            for gw in self.hm:
                if gw.name==srname:
                    break

        if self._srctype=='waterweb':
            gw = self._wwn.gwseries(srname)

        return gw





"""
Convert pd.DataFrame with coordinates to KML file

Created on 12-05-2013. Migrated to Acequia on 27-06-2020
Depends on package SimpleKML

"""

import os
import os.path
import warnings
from collections import OrderedDict

import pandas as pd
from pandas import DataFrame, Series
import numpy as np
import simplekml as skml

from .coordinate_conversion import CrdCon as CrdCon


sep = ','


class WpKml:
    """ Save Pandas DataFrame with waypoints in RD-coordinates to KML """

    ICONSHAPES = {'square' : 'http://maps.google.com/mapfiles/kml/shapes/placemark_square.png',
                  'circle' : 'http://maps.google.com/mapfiles/kml/shapes/placemark_circle.png',
                  'cross'  : 'http://maps.google.com/mapfiles/kml/shapes/cross-hairs.png'}

    COLORS = {
      'white'       : '#FFFFFF',
      'brightgreen' : '#33FF00',
      'mossgreen'   : '#CCFF00',
      'deepyellow'  : '#FFFF00',
      'darkyellow'  : '#FFCC00',
      'orange'      : '#FF6600',
      'lightblue'   : '#99CCFF',
      'deepblue'    : '#0000FF',
      'brightpurple': '#CC00CC',
      'darkpurple'  : '#6600CC',
      'pink'        : '#FF00FF',
      'brightred'   : '#FF0000'}


    DEFAULT_STYLEDICT = OrderedDict([
        ('1',{'iconshape':'square','iconscale':1.0,'iconcolor':'#0c7cac',
              'labelscale':0.7,'labelcolor':'#FFFFFF'}),
        ('2',{'iconshape':'square','iconscale':1.0,'iconcolor':'#ac0b35',
              'labelscale':0.7,'labelcolor':'#FFFFFF'}),
        ('3',{'iconshape':'square','iconscale':1.0,'iconcolor':'#9e00b5',
              'labelscale':0.7,'labelcolor':'#FFFFFF'}),
        ('4',{'iconshape':'square','iconscale':1.0,'iconcolor':'#54a910',
              'labelscale':0.7,'labelcolor':'#FFFFFF'}),
        ('5',{'iconshape':'square','iconscale':1.0,'iconcolor':'#f69601',
              'labelscale':0.7,'labelcolor':'#FFFFFF'}),
        ('6',{'iconshape':'square','iconscale':1.0,'iconcolor':'#0011ff',
              'labelscale':0.7,'labelcolor':'#FFFFFF'}),
        ('7',{'iconshape':'square','iconscale':1.0,'iconcolor':'#ff00c1',
              'labelscale':0.7,'labelcolor':'#FFFFFF'}),
        ('9',{'iconshape':'square','iconscale':1.0,'iconcolor':'#ffff00',
              'labelscale':0.0,'labelcolor':'#FFFFFF'}),
        ])


    missing_style = { 'iconshape':'cross',
                      'iconscale':1.0,
                      'iconcolor':'#FFFFFF',
                      'labelscale':0.7,
                      'labelcolor':'#FFFFFF'}


    def __repr__(self):
        return f'{self.__class__.__name__} (n={len(self.wplist)})'


    def __init__(self,wplist,label=None,xcoor=None,ycoor=None,
        styledict=None, stylecol=None):
        """Create KML file object from pd.DataFrame with waypoints

        Parameters
        ----------
        wplist : pd.DataFrame
            points with coordinates given in Standard Dutch Grid
        label : str
            columnname in wplist used for naming kml waypoints
        xcoor : float, optinal
            xcoordinate in RD
        ycoor : float, optional
            ycoordinate in RD
        styledic : dictionary, optional
            style definitions for each marker
        stylecol : str, optional
            column name for assigning styles

        Examples
        --------
        >>>sourcepath = '.\\testdata\\waypoints\\pmgloc.csv'
        >>>pmgloc = pd.read_csv(sourcepath)
        >>>pts = WpKml(pmgloc,label="nitgcode",xcoor="xcr",ycoor="ycr")
        >>>outdir = '.\\output\\kml\\'
        >>>pts.writekml(f'{outdir}pmv2019-noformatting.kml')

        >>>styledict_truefalse = {
                "true":
                    {"iconshape":"circle","iconscale":1.0,
                     "iconcolor":"#0000FF","labelscale":0.7,
                     "labelcolor":"FFFFFF"}),
                {"false":
                    {"iconshape":"circle","iconscale":1.0,
                     "iconcolor":"#FF00FF","labelscale":0.7,
                     "labelcolor":"FFFFFF"},}
        >>>pts = WpKml(pmgloc,label="nitgcode",xcoor="xcr",ycoor="ycr",
                       styledict=styledict_truefalse, stylecol="ispmg")
        >>>pts.writekml(f'{outdir}pmv-styledict-and-stylecol.kml')

        Notes
        -----
        If no values for xcoor,ycoor are given, column names with
        coordinates will be inferred from wplist column names.
        
        """

        self.kml = skml.Kml()
        self.defaultstyle = self._basestyle()

        if not isinstance(wplist, pd.DataFrame):
            raise TypeError('Variable wplist must be a pandas dataframe.')        

        if wplist.empty:
            raise ValueError('Variable wplist cannot be empty.')

        self.wplist = wplist.copy()
        self.wplist.columns = self.wplist.columns.str.lower()

        # set coordinates
        self.xcoor,self.ycoor = self._coordinate_columns(xcoor,ycoor)

        notlon = 'lon' not in self.wplist.keys()
        notlat = 'lat' not in self.wplist.keys()
        if notlon or notlat:
            self.wplist['lon'],self.wplist['lat'] = self.wgs84_coords

        # set label column name
        if label is None:
            self.label = self.wplist.keys()[0]

        if label.lower() in self.wplist.keys():
            self.label = label.lower()
        else:
            firstcol = list(self.wplist)[0]
            self.label = firstcol
            warnings.warn((f'Value {label} for label column not found '
                f'in {self.wplist} column names. Column {firstcol} is ',
                f'used for KML labels.'))

        if stylecol is not None:
            stylecol = stylecol.lower()

        # set kml waypoint styles
        # TODO: Test function _validate_styledict
        self.styledict, self.stylecol = self._validate_styledict(styledict,stylecol)
        self.pointstyles = OrderedDict()
        for iconkey,icondef in self.styledict.items():
            basestyle = self._basestyle()
            self.pointstyles[iconkey] = self._changestyle(basestyle,icondef)


    def _coordinate_columns(self,xcoor=None,ycoor=None):
        """Return names of columns with xcoor,ycoor """

        if xcoor is None:
            xlist = ['xcoor','xrd','x']
            for x in xlist:
                if x in self.wplist.keys(): 
                    xcol = x
        else:
            if xcoor.lower() in self.wplist.keys(): 
                xcol = xcoor.lower()
            else:
                msg = f'No column with xcoor value available.'
                raise ValueError(msg)

        if ycoor is None:
            ylist = ['ycoor','yrd','y']
            for y in ylist:
                if y in self.wplist.keys(): 
                    yxol = y
        else:
            if ycoor.lower() in self.wplist.keys():
                ycol = ycoor.lower()
            else:
                msg = f'No column with ycoor value available.'
                raise ValueError(msg)

        return xcol,ycol

    @property
    def wgs84_coords(self):
        """Return lon and lat as ps.Series

        Returns
        -------
        lon,lat
        """
        crdcon = CrdCon()
        lon = self.wplist.apply(
                lambda x: crdcon.RDtoWGS84(float(x[self.xcoor]),
                float(x[self.ycoor]))['Lon'],axis=1)
        lat = self.wplist.apply(
                lambda x: crdcon.RDtoWGS84(float(x[self.xcoor]),
                float(x[self.ycoor]))['Lat'],axis=1)
        return lon,lat

    def _validate_styledict(self,styledict,stylecol):
        """ validate styledict and stylecol """

        if (stylecol is not None) and (styledict is not None):

            # fill nans in column stylecol
            nanrows = self.wplist[self.wplist[stylecol].isnull()]
            if len(nanrows) !=0:
                fillnans = self.wplist[stylecol].replace(
                            np.nan,'missing',inplace=False)
                self.wplist[stylecol] = fillnans
                styledict['missing'] = self.missing_style


            # check styledict for missing keys
            # change stylecol to None if keys are missing from styledict
            missing = list(set(self.wplist[stylecol].unique()) 
                           - set(styledict.keys()))
            if len(missing)!=0: #not all(elem in wplabels for elem in styledict.keys()):               
                msg = ''.join([
                        f'Custom style dictionary styledict is missing ',
                        f'labels from waypoint column {stylecol}. ',
                        f'No formatting applied.\\n',
                        f'Missing labels are: {missing}'])
                warnings.warn(msg,UserWarning)
                stylecol = None


        if (stylecol is not None) and (styledict is None):

            wplabels = self.wplist[stylecol].unique()
            defaultkeys = list(self.DEFAULT_STYLEDICT.keys())
            if len(wplabels) <= len(defaultkeys):
                newstyles = []
                for i,wplabel in enumerate(wplabels):
                    newstyles.append((wplabel,self.DEFAULT_STYLEDICT[defaultkeys[i]]))
                styledict = OrderedDict(newstyles)
                #self.pointstyles = self._styles_from_dict(self.styledict)

            else:
                lenwp = len(wplabels)
                lenkeys = len(defaultkeys)
                message = ''.join(
                  [f'\nTo many unique values in waypoint stylecolumn '
                   f'{stylecol}.\n',
                   f'Stylecolumn {stylecol} contains {lenwp} values.\n',
                   f'Default style definitions contains {lenkeys} values.\n',
                   f'No formatting of waypoints applied.\n',])
                warnings.warn(message,UserWarning)
                styledict = {'default' : self.DEFAULT_STYLEDICT['1']}
                stylecol = None


        if (stylecol is None) and (styledict is not None):
            styledict = next(iter(styledict.values())) # abitrary single element from styledict

        if (stylecol is None) and (styledict is None):
                styledict = {'default' : self.DEFAULT_STYLEDICT['1']}


        """
        elif ((self.stylecol is None) and isinstance(self.styledict,dict)):
        if (stylecol is None) and (styledict is not None):
        if isinstance(styledict,dict):
            self.pointstyles = self._styles_from_dict(styledict)
        """
        return styledict,stylecol


    def _styles_from_dict(self,icondict):
        """Return OrderedDict of KML pointstyles with keys from icondict
        as point style names 

        Parameters
        ----------
        icondict : dict
            valid iconstyle dictionary


        Returns
        -------
        Ordereddict with KML pointstyles

        """
        pointstyles = OrderedDict()
        for iconkey,icondef in icondict.items():
            pointstyles[iconkey] = self._changestyle(
                                   kmlstyle=self._basestyle(),
                                   icondef=icondef)
        return pointstyles


    def _basestyle(self,icondict=None):
        """ Return default KML icon style """

        basestyle = skml.Style()
        basestyle.labelstyle.color = skml.Color.hex(self.COLORS['white'])
        basestyle.labelstyle.scale = 0.7
        basestyle.iconstyle.icon.href = self.ICONSHAPES['square']
        basestyle.iconstyle.scale = 0.9
        basestyle.iconstyle.color= skml.Color.hex('33FF33')

        # only show desciption in balloon
        basestyle.balloonstyle.text = r'$[description]' 

        if icondict is not None:
            basestyle = self._changestyle(kmlstyle=basestyle,
                                          icondef=icondict)

        return basestyle


    def _changestyle(self,kmlstyle=None,icondef={}):
        """ Change existing kml style with definitions given in dictionary

        Parameters
        ----------
        kmlstyle : skml.style
            existing skml style object
        icondef : dict
            dictionary with items and values to change

        Returns
        -------
        KML style object

        """

        if not isinstance(kmlstyle,skml.styleselector.Style):
            msg = f'Variable {kmlstyle} must be a simplekml style object.'
            raise TypeError(msg)

        if not isinstance(icondef,dict):
            msg = f'Variable {icondef} must be a dictionary not {type(icondef)}.'
            raise TypeError(msg)

        for itemkey in icondef.keys():

            if itemkey=='iconshape':
                icon_name = icondef['iconshape']
                kmlstyle.iconstyle.icon.href = self.ICONSHAPES[icon_name]

            elif itemkey=='iconscale':
                kmlstyle.iconstyle.scale = icondef['iconscale']

            elif itemkey=='iconcolor':
                mycolor = icondef['iconcolor']
                if mycolor.startswith('#'): 
                    mycolor = mycolor[1:]               
                kmlstyle.iconstyle.color= skml.Color.hex(mycolor)

            elif itemkey=='labelcolor':
                labelcolor = icondef['labelcolor']
                if labelcolor.startswith('#'): 
                    mycolor = mycolor[1:]               
                kmlstyle.labelstyle.color = skml.Color.hex(labelcolor)

            elif itemkey=='labelscale':
                kmlstyle.labelstyle.scale = float(icondef['labelscale'])

        return kmlstyle


    def _addcustomstyle(self,name,colorname=None,hexcolor=None,
        iconshape=None,styledic=None):
        """Add new custom KML waypoint style to customstyles"""

        newstyle = self.basestyle()

        if isinstance(styledic,dict):

            newstyle.iconstyle.icon.href = self.icons[styledic['iconshape']]
            newstyle.iconstyle.scale = styledic['iconscale']

            mycolor = styledic['iconcolor']
            if mycolor.startswith('#'): 
                mycolor = mycolor[1:]               
            newstyle.iconstyle.color= skml.Color.hex(mycolor)

            labelcolor = styledic['labelcolor']
            if labelcolor.startswith('#'): 
                mycolor = mycolor[1:]               
            newstyle.labelstyle.color = skml.Color.hex(labelcolor)
            newstyle.labelstyle.scale = float(styledic['labelscale'])

        if colorname is not None:

            if colorname in self.COLORS.keys(): 
                hexcolor = self.COLORS[colorname]

            if hexcolor.startswith('#'): 
                hexcolor = int(hexcolor[1:])

            newstyle.iconstyle.color= skml.Color.hex(hexcolor)

        if iconshape is not None:

            if iconshape in self.icons.keys(): 
                iconshape = self.icons[iconshape]
            else: 
                iconshape = self.icons['square']
            newstyle.iconstyle.icon.href = iconshape

        self.customstyles[name] = newstyle
        return self.customstyles[name]


    def writekml(self,kmlfile,metadata=True,cols=None):
        """ Write list of waypoints in RD to KML

        Parameters
        ----------
        kmlname : str
            name of kml file
        metadata : True,False, optional
            add metadata in bollontext (True) or not (False)
        cols : list of str, optional
            list of column names to add to balloontext of waypoints
        """
        #TODO: check if kmlfile path exists

        # Make list of fields to write to balloontext
        ##self.ballooncols = cols


        for idx, wp in self.wplist.iterrows():
   
            pnt = self.kml.newpoint(name=wp[self.label], 
                        coords=[(wp['lon'],wp['lat'])])  # lon, lat optional height
            pnt.snippet.maxlines = 0
            pointname = r'<font color='#000' size='+1'><strong>'+wp[self.label]+'</strong></font><br>'

            # assign waypoint style
            if (isinstance(self.stylecol,str) and isinstance(self.styledict,dict)):
                pnt.style = self.pointstyles[wp[self.stylecol]]
            else:
                pnt.style = self._basestyle(icondict=self.styledict)

            if metadata is True:

                if cols is None:
                    cols = list(wp.index)
                else:
                    cols = [x.lower() for x in cols]
                fields = zip(wp[cols].index, wp[cols].values)

                for ifield,field in enumerate(fields):
                    if ifield==0: 
                        tekst = field[0]+': '+str(field[1])+'<br>'
                    else: 
                        tekst = tekst + field[0]+': '+str(field[1])+'<br>'
            
                # save balloontext to kml
                pnt.description = tekst

        self.kml.save(kmlfile)

        return self.kml





import pathlib
import warnings
import numpy as np
from pandas import Series,DataFrame
import pandas as pd
import geopandas as gpd
##from shapely.geometry import Point
from ..gwseries import GwSeries
from ..geo.waypoint_kml import WpKml


class WaterWeb:
    """
    Manage WaterWeb dataset

    Constructor
    -----------
    from_csv
        Read waterweb csv export file and return WaterWeb object.

    """
    _column_mapping = {
        'Lokatie':'sunloc',
        'SUN-kode':'sunsr',
        'NITG-kode':'nitgsr',
         #'OLGA-kode':'olga', #VERVALLEN
        'BROID':'broid',
        'DERDEN-kode':'derden',
        'X coordinaat':'xcr',
        'Y coordinaat':'ycr',
        'NAP hoogte bovenkant peilbuis':'mpnap',
        'Hoogte maaiveld tov NAP':'mvnap',
        'Hoogte maaiveld tov Nulpunt':'mvmp',
        'Hoogte maaiveld tov maaiveld':'mvmv',
        'NAP hoogte bovenkant filter':'filtopnap',
        'NAP hoogte onderkant filter':'filbotnap',
        'Peilmoment':'datetime',
        'Peilstand':'peilcmmp',
        'Peilstand in Meters':'peilmmp',
        'Peilstand tov NAP':'peilcmnap',
        'Peilstand tov NAP in Meters':'peilmnap',
        'Peilstand tov maaiveld':'peilcmmv',
        'Peilstand tov maaiveld in Meters':'peilmmv',
        'Peilkode':'peilcode',
        'Opmerking bij peiling':'peilopm'
        }

    """
    _typedict = {
        'Lokatie':str,
        'SUN-kode':str,
        'NITG-kode':str,
        'BROID':str,
        ##'OLGA-kode':str,
        'DERDEN-kode':str,
        'X coordinaat':'float64',
        'Y coordinaat':'float64',
        'NAP hoogte bovenkant peilbuis':'int64',
        'Hoogte maaiveld tov NAP':'int64',
        'Hoogte maaiveld tov Nulpunt':'int64',
        'Hoogte maaiveld tov maaiveld':'int64',
        'NAP hoogte bovenkant filter':'int64',
        'NAP hoogte onderkant filter':'int64',
        'Peilmoment':str,
        'Peilstand':'int64',
        'Peilstand in Meters':'float64',
        'Peilstand tov NAP':'int64',
        'Peilstand tov NAP in Meters':'float64',
        'Peilstand tov maaiveld':'int64',
        'Peilstand tov maaiveld in Meters':'float64',
        'Peilkode':str,
        'Opmerking bij peiling':str,
        }
    """
    
    _locprops_cols = ['sunloc','sunsr','nitgsr','broid','derden',
        'xcr','ycr',]

    _tubeprops_cols = ['mpnap','mvnap','filtopnap','filbotnap']

    _peilprops_cols = ['datetime','peilcmmp','peilcode','peilopm']

    _locprops_mapping = {
        'locname':'sunloc','alias':'nitgsr',
        'xcr':'xcr','ycr':'ycr'
        }

    _tubeprops_mapping = {
        'startdate':'datetime','mplevel':'mpnap',
        'filtop':'filtopnap','filbot':'filbotnap',
        'surfacelevel':'mvnap'
        }

    _levels_mapping = {
        'headdatetime':'datetime', 'headmp':'peilcmmp', 
        'headnote':'peilcode','remarks':'peilopm'}

    _reflevels = [
        'datum','surface','mp',
        ]

    MEASUREMENT_TYPES = ['B','S','L','P']

    KMLSTYLES = {
        'B':
            {'iconshape':'circle', 'iconscale':1.2,
             'iconcolor':'#0000FF','lobalescale':0.7,
             'labelcolor':'FFFFFF'},                
        'S':
            {'iconshape':'circle', 'iconscale':1.2,
             'iconcolor':'#1ca3ec','lobalescale':0.7,
             'labelcolor':'FFFFFF'},                
        'L':
            {'iconshape':'circle', 'iconscale':1.2,
             'iconcolor':'#2389da','lobalescale':0.7,
             'labelcolor':'FFFFFF'},                
        'P':
            {'iconshape':'circle', 'iconscale':1.2,
             'iconcolor':'#5abcd8','lobalescale':0.7,
             'labelcolor':'FFFFFF'},                
        }


    def __init__(self,fpath=None,data=None,network=None):

        self._fpath = fpath
        self._network = network
        self._rawdata = data
        self.data = data

        if self._network is None:
            self._network = 'Onbekend Waterweb meetnet'

        if self.data is not None:

            if not isinstance(self.data,pd.DataFrame):
                raise ValueError((f'{self.data} is not a valid Pandas '
                    f'DataFrame.'))

            # rename columns in data
            self.data = self.data.rename(columns=self._column_mapping)

            # Remove measurements without reference level.
            # Note:
            # When thera are measurements with dates earlier than the
            # first date with technical tube data, the waterweb csv 
            # export starts with measurments without technical data.
            # These are removed and a user warnimg is given.
            hasnoref = self.data['mpnap'].isnull()
            if len(self.data[hasnoref])!=0:
                warnings.warn((f'{len(self.data[hasnoref])} measurements '
                    f'were removed from {self.networkname}'))
                self.data = self.data[~hasnoref].copy()

    def __repr__(self):
        return (f'{self._network} (n={self.__len__()})')


    def __len__(self):
        return len(self.srnames)


    @classmethod
    def from_csv(cls,fpath,network=None):
        """ 
        Read waterweb csv network file and return new WaterWeb object

        Parameters
        ----------
        filepath : str
            path to waterweb csv export file

        networkname : str, optional
            name of network

        Returns
        -------
        WaterWebNetwork object
        ...
        """
        try:
            data = pd.read_csv(fpath,sep=';',decimal=',')
        except FileNotFoundError as err:
            raise FileNotFoundError(f'Invalid filepath for WaterWeb csv file: "{fpath}"')

        if network is None:
            network = pathlib.Path(fpath).stem

        #check for missing columns
        missing_columns = []
        for col in WaterWeb._column_mapping.keys():
            if col not in list(data):
                missing_columns.append(col)
        if missing_columns:
            warnings.warn((f'Missing columns in WaterWeb csv file: '
                f'{missing_columns}'))

        # check for unknown columns
        unknown_columns = []
        for col in list(data):
            if col not in WaterWeb._column_mapping.keys():
                unknown_columns.append(col)
        if unknown_columns:
            warnings.warn((f'Unknown columns in WaterWeb csv file: '
                f'{unknown_columns}.'))

        ## change column types
        ##data = data.astype(dtype=cls._typedict)

        # change column contents
        data['Peilmoment'] = pd.to_datetime(data['Peilmoment'])
        data['NITG-kode'] = data['NITG-kode'].apply(
            lambda x:x[:8]+"_"+x[-3:].lstrip('0') if not pd.isnull(x) else np.nan)

        return cls(fpath=fpath,data=data,network=network)


    @property
    def srnames(self):
        """Return list of series names"""
        if self.data is not None:
            return list(self.data['sunsr'].unique())
        else:
            return []

    def get_type(self,srname=None):
        """Return kind of measurement type series

        Parameters
        ----------
        srname : str, optional
            series name, if not given, type of all series is returned

        Returns
        -------
        str, numpy array of str """

        srnames = self.srnames
        sr = Series(data=srnames,index=srnames,name='seriestype')
        sr = sr.apply(lambda x:x[8])

        if srname is not None:
            return sr[srname]

        return sr

    @property
    def measurement_types(self):
        """Return table of measurement type counts."""
        srtypelist = []
        for name in self.srnames:
            srtypelist.append(self.get_type(name))
        tbl = pd.Series(srtypelist).value_counts()
        tbl.name = self.networkname
        return tbl


    def get_locname(self,srname):
        """Return location name for given series

        Parameters
        ----------
        srname : str
            name of series to return """

        return self.get_locprops(srname)['sunloc']


    def get_filname(self,srname):
        """Return filter name for given series

        Parameters
        ----------
        srname : str
            name of series to return 

        Notes
        -----
        WaterWeb uses the DINO-SUN convention of only explicitly naming
        filters if more than one filter is present."""

        filname = self.get_locprops(srname)['sunsr']
        if filname[-1].isalpha():
            return filname[-1]
        return ''


    @property
    def networkname(self):
        """Return network name

        Parameters
        ----------
        name : str, optional
            name of measurement network """
        return self._network

    @networkname.setter
    def networkname(self,name):
        self._network = name

    def get_locprops(self,srname):
        """Return series location properties

        Parameters
        ----------
        sunsr : str
            name of series to return

        Return
        ------
        pd.Series """

        data =self.data[self.data['sunsr']==srname]
        lastrow = data.iloc[-1,:]

        sr = Series(index=self._locprops_cols,dtype='object',
            name=srname)
        for col in self._locprops_cols:
            sr[col] = lastrow[col]

        return sr


    def get_tubeprops(self,srname):
        """Return welltube properties

        Parameters
        ----------
        sunsr : str
            name of series

        Returns
        -------
        pd.DataFrame """

        data =self.data[self.data['sunsr']==srname]
        data = data.drop_duplicates(subset=self._tubeprops_cols,
            keep='first')
        data = data[['datetime']+self._tubeprops_cols]
        data = data.reset_index(drop=True)

        return data

    def get_levels(self,srname,ref='mp'):
        """Return measured levels

        Parameters
        ----------
        srname : str
            name of series
        ref : {'mp','datum',' surface'}, default 'mp' 

        Returns
        -------
        pd.Series """

        if ref not in self._reflevels:
            warnings.warn((f'{ref} not in {self._references}. '),
                (f'reference is set to "datum".')) 
            ref = 'datum'
        if ref=='mp':
            col = 'peilcmmp'
        if ref=='datum':
            col = 'peilmnap'
        if ref=='mv':
            col = 'peilcmmv'

        data =self.data[self.data['sunsr']==srname]
        data = data[[col,'datetime']]
        data = data.set_index('datetime',drop=True).squeeze()
        data.name = self.get_locname(srname)
        data.index.name = 'datetime' 
        data = data/100.

        return data


    def get_gwseries(self,srname):
        """Return gwseries obect for one series

        Parameters
        ----------
        srname : str
            name of series to return

        Returns
        -------
        acequia.GwSeries

        """

        gw = GwSeries()

        # locprops
        locprops = self.get_locprops(srname)
        for gwprop in list(gw._locprops.index):
            if gwprop not in self._locprops_mapping.keys():
                continue
            wwnprop = self._locprops_mapping[gwprop]
            gw._locprops[gwprop] = locprops[wwnprop]

        gw._locprops['filname'] = self.get_filname(srname)
        gw._locprops['height_datum'] = 'mNAP'
        gw._locprops['grid_reference'] = 'RD'

        # tubeprops
        tubeprops = self.get_tubeprops(srname)
        for gwprop in list(gw._tubeprops):
            if gwprop not in self._tubeprops_mapping.keys():
                continue
            wwnprop = self._tubeprops_mapping[gwprop]
            gw._tubeprops[gwprop] = tubeprops[wwnprop].values
            if gwprop in gw._tubeprops_numcols:
                gw._tubeprops[gwprop] = gw._tubeprops[gwprop]/100.

        #levels
        levels =self.data[self.data['sunsr']==srname]
        levels = levels[self._peilprops_cols]
        for gwprop in list(gw._headprops_names):
            if gwprop not in self._levels_mapping.keys():
                continue
            wwnprop = self._levels_mapping[gwprop]
            gw._heads[gwprop] = levels[wwnprop].values
        gw._heads['headmp'] = gw._heads['headmp']/100.

        return gw

    @property
    def locations(self):
        """Return locations as GeoDataFrame."""

        propslist = []
        for srname in self.srnames:
            sr = self.get_locprops(srname)
            sr['name'] = sr.name
            sr['mptype'] = self.get_type(srname)
            sr['network'] = self.networkname
            propslist.append(DataFrame(sr).T)
        locprops = pd.concat(propslist,ignore_index=True)

        # from series to locations
        locprops = locprops.drop_duplicates(subset=['sunloc'], keep='first')
        nitg = locprops['nitgsr'].apply(lambda x:x.split('_')[0] if 
            not pd.isnull(x) else np.nan)
        locprops.insert(2,'nitgcode',nitg)
        locprops = locprops.drop(columns=['sunsr','nitgsr','name'])

        # add waypoint label column
        labels = locprops['sunloc'].str[8]+locprops['sunloc'].str[9:].str.lstrip('0')
        locprops.insert(0,'label',labels)

        gdf = gpd.GeoDataFrame(
            locprops, geometry=gpd.points_from_xy(
            locprops['xcr'], locprops['ycr'], crs='EPSG:28992'))

        return gdf

    def to_kml(self,filepath):
        """Save locations to KML file.
        
        Parameters
        ----------
        filepath : str
            Valid path to output file.

        Returns
        -------
        wp : WpKml
        """

        locs = self.locations
        colnames = [col for col in list(locs) if col not in ['geometry']]
        wp = WpKml(locs[colnames],label='label',xcoor='xcr',ycoor='ycr',
            styledict=self.KMLSTYLES,stylecol='mptype')
        wp.writekml(filepath)
        return wp

    def to_gpx(self,filepath):
        """Save locations to GPX waypoints file.
        
        Parameters
        ----------
        filepath :str
            Valid filepath for saving GPX file.

        Returns
        -------
        GeoDataFrame
            GPX data that have been saved.
        """
        # source for this solution:
        # https://github.com/geopandas/geopandas/issues/684
        # jose1911 suggestes adding all requeierd files to dataframe en 

        locs = self.locations.copy()
        locs = locs.to_crs(4326)

        # adding columns that are propably expected by apps that import
        # gpx waypoint files.
        locs['name']=locs['label']
        locs['ele']=0
        locs['magvar']=0
        locs['time']='2019-08-02T14:17:50Z'
        locs['geoidheight'] = 0
        colnames = ['geometry', 'ele', 'time', 'magvar', 'geoidheight', 'name']
        
        # saving gpx file
        if not filepath.endswith('.gpx'):
            filepath = f'{filepath}.gpx'
        locs[colnames].to_file(filepath,'GPX')
        return locs[colnames]

    def iteritems(self):
        """Iterate over all series and return gwseries object."""
        for srname in self.srnames:
            yield self.get_gwseries(srname)



"""
Read .gpx tracklog files created by GPS receivers. 
Supports GPX v1.0 and v1.1
"""

from pathlib import Path
from pandas import DataFrame, Series
import pandas as pd
from lxml import etree

class GpxTree:
    """XML tree with GPS receiver tracklog file."""

    def __init__(self,tree, filepath=None):
        """
        Parameters
        ----------
        tree : ElementTree
            XML tree with data from .gpx file.
        flepath : str
            Filepath to .gpx sourcefile, optional
        """
        self._tree = tree
        self._filepath = filepath

        try:
            self._root = self._tree.getroot()
            self._ns = self._root.nsmap
        except AttributeError as err:
            clsname = tree.__class__.__name__
            raise AttributeError((f'Invalid gpx tree of class {clsname}'))

        # dict of nodes under root
        self._nodes_under_root = {}
        for node in self._root.iterchildren():            
            tag = node.tag
            key = tag.split('}')[1] # drop namespace
            self._nodes_under_root[key] = node

    def __repr__(self):
        try:
            return f"{self.__class__.__name__} ('{self.meta['desc']}')"
        except KeyError as err:
            return f"{self.__class__.__name__} (no description)"

    @classmethod
    def from_file(cls,fpath):
        """Read gpx from file.
        
        Parameters
        ----------
        fpath : str
            Valid filepath to .gpx file.

        Returns
        -------
        GpxTree
        """
        try:
            open(Path(fpath)) #test open to get usefull error message
            tree = etree.parse(fpath)
            
        except FileNotFoundError as error:
            raise FileNotFoundError(error) from None
        except OSError as error:
            raise OSError(error) from None

        return cls(tree,filepath=fpath)

    def _get_meta(self):
        """Extract tracklog metadata."""
        
        # metadata from root node
        self._meta = {}
        try:
            self._meta['gpx_version'] = self._root.attrib['version']
            self._meta['creator'] = self._root.attrib['creator']
        except (AttributeError,KeyError) as err:
            self._meta['gpx_version'] = None
            self._meta['creator'] = None
            print(err)

        # metadata from root child nodes
        if 'metadata' in self._nodes_under_root.keys(): #GPX v1.1
            node = self._nodes_under_root['metadata']
            for child in node:
                key = child.tag.split('}')[1] #drop namespace
                value = child.text
                if not value is None:
                    self._meta[key] = value
        else: #GPX v1.0
            for key in ['name','desc','time','keywords']:
                if key in self._nodes_under_root.keys():
                    value = self._nodes_under_root[key].text
                    if value is not None:
                        self._meta[key] = value
        return self._meta

    def _get_waypoints(self):
        """Extract waypoints from xml tree."""

        # extract waypoints from tree
        waypoints = []
        for wpnode in self._root.findall(f'.//wpt',self._ns):
            wp = {}
            wp['lat'] = wpnode.attrib['lat']
            wp['lon'] = wpnode.attrib['lon']
            for child in wpnode.iterchildren():
                key = child.tag.split('}')[1] #drop namespace
                wp[key] = child.text
            waypoints.append(wp)

        return DataFrame(waypoints)


    def _get_trackpoints(self):
        """Extract trackpoints from xml tree."""

        # iterate over track nodes
        trackpoints = []
        for tracknode in self._root.findall(f'.//trk',self._ns):
            trackname = tracknode.find(f'.//name',self._ns).text

            # iterate over track segments
            for tracksegment in tracknode.findall(f'.//trkseg',self._ns):
                segmentid = 0

                # iterate over segment trackpoints 
                for trackpoint in tracksegment.iterchildren():
                    tp = {}
                    tp['trackname'] = trackname
                    tp['segmentid'] = str(segmentid)
                    
                    # attribute of trackpoint
                    tp['lat'] = trackpoint.attrib['lat']
                    tp['lon'] = trackpoint.attrib['lon']
                    
                    # children of trackpoint
                    for child in trackpoint.iterchildren():
                        key = child.tag.split('}')[1] #drop namespace
                        tp[key] = child.text

                    trackpoints.append(tp)

                segmentid+=1

        return DataFrame(trackpoints)

    @property
    def meta(self):
        """Return .gpx metadata as DataFrame."""
        return self._get_meta()

    @property
    def bounds(self):
        """Return boundary of tracklog as Dict."""
        bounds_node = self._root.find(f'.//bounds',self._ns)
        try:
            return bounds_node.attrib
        except AttributeError as err:
            return {'minlat':None, 'minlon':None, 'maxlat':None,'maxlon':None}
        return ''

    @property
    def trackpoints(self):
        """Return .gpx trackpoints as DataFrame."""
        return self._get_trackpoints()

    @property
    def waypoints(self):
        """Return .gpx waypoints as DataFrame."""
        data = self._get_waypoints()
        return data

    @property
    def contents(self):
        """Return names of tags below xml root as list."""
        return list(self._nodes_under_root.keys())



""" This module contains a class Quantiles that calculates quantiles
of measured heads for hydrological years.

"""

from datetime import datetime
import datetime as dt
import math
import warnings
import numpy as np
from pandas import Series, DataFrame
import pandas as pd
import matplotlib.pyplot as plt

from ..gwseries import GwSeries
from .utils import hydroyear

class Quantiles:
    """Calculate cumulative frequencies of groundwater head series."""

    N14 = 18

    def __init__(self, heads, srname=None, headsref='surface'):
        """
        Calculate quantiles from series of measured heads.

        Parameters
        ----------
        heads : pd.Series, aq.GwSeries
            Timeseries with groundwater head measurements.

        srname : str
            User defined series name.

        headsref : {'datum','surface'}, default 'surface'
            Reference level for measurements.

        """

        if isinstance(heads,GwSeries):
            ts = heads.heads(ref=headsref)
            if srname is None:
                srname = heads.name()

        elif isinstance(heads,pd.Series):
            ts = heads
            if srname is None: 
                srname = heads.name
                
        elif isinstance(heads,DataFrame):
            ts = heads.iloc[:,0].squeeze()
            if srname is None: 
                srname = list(heads)[0]

        self._heads = heads
        self.headsref = headsref
        self.ts = ts
        self.srname = srname


    def __repr__(self):
        return f'{self.__class__.__name__}({self.srname})'


    def get_quantiles(self,unit='days',step=None):
        """Return table with quantiles for each hydrological year.
        
        Parameters
        ----------
        unit : {'days','quantiles'}, default 'days'
            Unit of quantile boundary classes.
        step : float or int
            Quantile class division steps. For unit days an integer 
            between 0 and 366, for unit quantiles a fraction between 
            0 and 1.

        Returns
        -------
        DataFrame, table with cumulative frequencies.
        """

        # set values for cumulative frequency table
        if unit=='days':
            if step is None: 
                step = 30
            if not 0 < step < 366:
                warnings.warn((f'For unit days, step must be an '
                    f'integer between 0 and 366, not {step}. '
                    f'Default value of 30 will be used.'))
                step = 30
            self.days = list(range(0,366,step))
            self.qt = [x/365 for x in self.days]
            self.qtlabels = [str(x) for x in self.days]

        elif unit=='quantiles':
            if step is None: 
                step = 0.1
            if not 0 < step < 1:
                warnings.warn((f'For unit quantiles, step must be a '
                    f'fraction between 0 and 1, not {step}. '
                    f'Default value of 0.1 will be used.'))
                step = 0.1

            self.qt = np.arange(0,1+step,step) # list of quantiles
            self.qt = self.qt[self.qt<=1] #delete qts >1.0
            self.qtlabels = ['p'+str(int(x*100)) for x in self.qt]
            self.days = [int(x*365) for x in self.qt]

        else:
            raise ValueError((f'Value for parameter unit must be "days" '
                f'or "quantiles", not {unit}'))

        # create empty table with hydroyears and percentiles
        hydroyears = hydroyear(self.ts)
        unique_years = np.unique(hydroyears)
        quantiles = pd.DataFrame(index=unique_years,columns=self.qtlabels)

        # calculate quantiles
        grp = self.ts.groupby(hydroyears)
        for i,(name,quantile) in enumerate(zip(self.qtlabels,self.qt)):
            if self.headsref=='datum':
                quantiles[name] = grp.quantile(quantile).round(2)
            elif self.headsref=='surface':
                quantiles[name] = grp.quantile(quantile)*100
                quantiles[name] = quantiles[name].apply(
                    lambda x:math.floor(x) if not np.isnan(x) else x)
                ##).round(0).astype(int)
        return quantiles

    def get_summary(self,unit='days',step=None, decimals=2):
        """Return table with summary of quantile statistics.
        
        Parameters
        ----------
        unit : {'days','quantiles'}, default 'days'
            Unit of quantile boundary classes.
        step : float or int
            Quantile class division steps. For unit days an integer 
            between 0 and 366, for unit quantiles a fraction between 
            0 and 1.
        decimals : int, default 2
            Round results to specified number of decimals.

        Returns
        -------
        DataFrame
        """
        qt = self.get_quantiles(unit=unit,step=step)

        decimals = 2
        q05 = qt.quantile(q=0.05, axis=0, numeric_only=True).round(decimals)
        q95 = qt.quantile(q=0.95, axis=0, numeric_only=True).round(decimals)
        mean = qt.mean().round(decimals)
        std = qt.std().round(decimals)
        count = qt.count()

        summary = pd.concat({'q05':q05,'q95':q95,'mean':mean,'std':std,'count':count},axis=1).T
        if self.headsref == 'surface': # round all values to zero decimals
            rows = []
            for rowlabel in summary.index:
                rows.append(summary.loc[rowlabel,:].apply(
                    lambda x:math.floor(x) if not np.isnan(x) else x))
            summary = DataFrame(rows)
            
        return summary

    def get_inundation(self, unit='days',step=5):
        """Return number of days with inundation as mean, min, max.
        
        Parameters
        ----------
        unit : {'days','quantiles'}, default 'days'
            Unit of quantile boundary classes.
        step : float or int, default 5
            Quantile class division steps. For unit days an integer 
            between 0 and 366, for unit quantiles a fraction between 
            0 and 1.

        Returns
        -------
        pandas.Series
        ...
        """

        def get_negative_days(sr):
            if sr[sr<0].empty:
                return 0
            return float(sr[sr<0].index[-1])

        summary = self.get_summary(unit=unit,step=step)
        sr = Series(data=[
            get_negative_days(summary.loc['mean',:]),
            get_negative_days(summary.loc['q95',:]),
            get_negative_days(summary.loc['q05',:]),
            ],index = ['mean','min','max'],name='inundationtime')

        if self.headsref == 'surface': # round all values to zero decimals
            sr = sr.apply(lambda x:math.floor(x) if not np.isnan(x) else x)

        return sr

    def get_lowest(self, unit='days',step=5):
        """Return lowest measured heads.
        
        Parameters
        ----------
        unit : {'days','quantiles'}, default 'days'
            Unit of quantile boundary classes.
        step : float or int, default 5
            Quantile class division steps. For unit days an integer 
            between 0 and 366, for unit quantiles a fraction between 
            0 and 1.

        Returns
        -------
        pandas.Series
        """

        summary = self.get_summary(unit=unit,step=step)
        lastcol = summary.columns[-1]
        sr = Series([
            summary.at['mean',lastcol],
            summary.at['q05',lastcol],
            summary.at['q95',lastcol],
            ], index = ['mean','min','max'],
            name='lowest')
        return sr



    def plot(self,unit='days',step=30,median=False,coloryears=None,boundyears=None,
        figpath=None,figtitle=None,ylim=None, ax=None):
        """Plot quantiles

        Parameters
        ----------
        unit : {'days','quantiles'}, default 'days'
            Unit of quantile boundary classes.
        step : float or int,default 30
            Quantile class division steps. For unit days an integer 
            between 0 and 366, for unit quantiles a fraction between 
            0 and 1.
        coloryears : list of int, optional
            List of years to plot in color
        boundyears : list, optional
            List of years to include in calculating boundaries for 
            filled surface.
        figpath : str, optional
            figure output path
        figtitle : str, optional
            figure title
        ylim : list, optional
            [yaxmin, yaxmax]
        ax : matplotlib.axes.Subplot, optional
            Ax to plot on.

        Returns
        -------
        matplotlib.pyplot.ax
        """

        quantiles = self.get_quantiles(unit=unit,step=step)

        if coloryears is None:
            coloryears = []

        if not isinstance(coloryears,list):
            warnings.warn((f'years of type {type(colorlist)} changed to ' 
                'empty list'))
            years = []

        if figtitle is None:
            figtitle = self.srname

        csurf = '#8ec7ff'
        clines = '#2f90f1' #'#979797' #
        cyears = '#b21564'
        cmedian = '#8b0000'

        if ax is None:
            fig,ax = plt.subplots(1,1)
            fig.set_figwidth(8)
            fig.set_figheight(4)

        x = self.qt
        reftbl = quantiles.copy()
        if boundyears is not None:
            idx = [x for x in reftbl.index.values if x not in boundyears]
            reftbl = reftbl.loc[idx,:]

        # reference based on quantiles
        upper = reftbl.quantile(0.05)*100
        lower = reftbl.quantile(0.95)*100

        # alternative reference
        #mean = reftbl.median(axis=0,skipna=True)
        #std = reftbl.std(axis=0,skipna=True)
        #upper = (mean+std)*100
        #lower = (mean-std)*100

        # alternative reference 2
        # leave out lowest line
        #upper = [reftbl[col].sort_values()[:-1].quantile(0.05)*100 for col in reftbl.columns]
        #lower = [reftbl[col].sort_values()[:-1].quantile(0.95)*100 for col in reftbl.columns]

        # draw colored surface with reference
        ax.fill_between(x, upper, lower, color=csurf) 
        
        # plot line for each year
        for year in quantiles.index:
            yvals = quantiles.loc[year,:].values * 100
            xvals = self.qt
            ax.plot(xvals,yvals,color=clines)

        # plot colored lines
        for year in quantiles.index:
            if year in coloryears:
                yvals = quantiles.loc[year,:].values * 100
                xvals = self.qt
                ax.plot(xvals,yvals,color=cyears)

        if median:
            yvals = quantiles.median().values * 100
            xvals = self.qt
            ax.plot(xvals,yvals,color=cmedian)

        # xax set ticks
        qtlen = len(self.qt)
        step = int(np.ceil((qtlen-2)/6))
        idx = list(range(0,qtlen,step))
        xticks = [self.qt[i] for i in idx]
        ax.set_xticks(xticks)

        #qlabels = list(quantiles)
        #qlables = ['0','90','180','270','360']
        qtlabels = [self.qtlabels[i] for i in idx]
        ax.set_xticklabels(qtlabels,fontsize=7.) ##self.qtlabels)
        if unit=='days':
            ax.set_xlabel('dagen/jaar') #, fontsize=15)
        else:
            ax.set_xlabel('percentiel') #, fontsize=15)
        ax.set_xlabel('')

        ax.set_xlim(1,0)
        ax.invert_xaxis()

        # format yax
        if self.headsref=='surface':
            ax.invert_yaxis()

        if ylim is not None:
            ax.set_ylim(ylim[0],ylim[1])

        yticklabels = [str(int(x)) for x in ax.get_yticks()]
        ax.set_yticklabels(yticklabels,fontsize=10.) 
        
        ax.set_ylabel('')

        ax.text(0.99, 0.97, figtitle, horizontalalignment='right',
                verticalalignment='top', transform=ax.transAxes,
                fontsize = 10.)

        if figpath is not None:
            plt.savefig(figpath,bbox_inches='tight')

        return ax





import pytest
from pandas import DataFrame
from acequia import brorest
from acequia import BroGldXml


@pytest.fixture
def gmwid():
    return 'GMW000000041145'

@pytest.fixture
def brogld():
    return 'GLD000000010138'
    
@pytest.fixture
def broinstantie():
    return '51048329'

def test_getareawellprops():

    circle = brorest.get_area_wellprops(lowerleft=None,upperright=None,
        center=(52.349968,7.064451),radius=0.5)
    assert isinstance(circle,DataFrame)
    assert not circle.empty

    rectangle = brorest.get_area_wellprops(lowerleft=(52.340333,6.865430),
        upperright=(52.347915,6.888625),center=None,radius=None)
    assert isinstance(rectangle,DataFrame)
    assert not rectangle.empty

def test_singlewellprops(gmwid):
    props = brorest.get_wellprops(gmwid)
    assert isinstance(props,DataFrame)
    assert not props.empty

def test_getwelltubes(gmwid):
    tubes = brorest.get_welltubes(gmwid)
    assert isinstance(tubes,DataFrame)
    assert not tubes.empty

def test_getputcode(gmwid):
    wellcode = brorest.get_putcode(gmwid)
    assert isinstance(wellcode,str)
    assert len(wellcode)!=0

def test_getlevels(brogld):
    levels = brorest.get_levels(brogld=brogld) #'GLD000000010138') #'GLD000000009881')
    assert isinstance(levels,BroGldXml)
    assert not levels.heads.empty

def test_getgmwcodes(broinstantie):
    gmwlist = brorest.get_gmw_codes('51048329')
    assert isinstance(gmwlist,list)
    assert len(gmwlist)!=0

def test_getgldcodes(broinstantie):
    gldlist = brorest.get_gld_codes('51048329')
    assert isinstance(gldlist,list)
    assert len(gldlist)!=0



"""
Module with GwFiles object that holds a list of groundwater series
source filepaths.
"""

import os
import os.path
import warnings
from pandas import Series, DataFrame
import pandas as pd

from ..gwseries import GwSeries

class GwFiles:
    """Collection of groundwater head source files.
    
    Files with groundwater head data are read from a sourcefolder
    and can be iterated over returning GwSeries objects.
    Data can also be written to output directories as json or csv files 
    with "to_json" or "to_csv" methods.
    
    Examples
    --------
    >>> srcdir = <directory with dinoloket sourcefiles>
    >>> gwf = GwFiles.from_dinocsv(srcdir)
    >>> for gw in gwf.iteritems():
    ...     print(gw)
    >>> gwf.to_json(<outputdir>)
    ...
    """

    FILETABLE_COLS = ['series', 'loc', 'fil', 'fname', 'fpath',]

    def __init__(self, filetable=None, source=None):
        """Create a GwFiles object with class methods from_dinocsv, 
        from_json or from_csv.
        
        Examples
        --------
        >>> srcdir = <valid path to directory with dinoloket sourcefiles>
        >>> gwf = GwFiles.from_dinocsv(srcdir)
        ...
        """

        # test validity of filetbl
        if not isinstance(filetable, DataFrame):
            raise ValueError((f'Argument filetable must be {type(DataFrame())} '
                f'not {type(filetable)}'))

        missing_cols = [col for col in self.FILETABLE_COLS if col not in filetable.columns]
        if missing_cols:
            raise ValueError(('Filetable is missing required columns: '
                f'{missing_cols}.'))

        self._ftb = filetable
        self._source = source

    def __len__(self):    
        return len(self._ftb)

    def __repr__(self):
        return (f'{self.__class__.__name__} (n={len(self._ftb)})')

    @classmethod
    def from_dinocsv(cls,srcdir,loclist=None):
        """Create GwFiles object from folder with DinoLoket sourcefiles.
        
        Parameters
        ----------
        srcdir : str
            Path to directory with Dinoloket csv sourcefiles.
        loclist : list, optional
            List of strings with valid location names to restrict
            number of files read from srcdir.
        """

        if not os.path.isdir(srcdir):
            raise ValueError(f'Directory {srcdir} does not exist')

        # table of filenames
        filenames = [fname for fname in os.listdir(srcdir) if fname.endswith('1.csv')]
        filetbl = pd.DataFrame({"fname":filenames})

        # add columns to filetbl
        filetbl["fpath"]= filetbl["fname"].apply(lambda x:srcdir+x)
        filetbl.insert(0,"loc",filetbl["fname"].apply(lambda x:x[0:8]))
        filetbl.insert(1,"fil",filetbl["fname"].apply(lambda x:x[8:11].lstrip("0")))
        filetbl.insert(0,"series",filetbl["loc"]+"_"+filetbl['fil'])

        if loclist is not None:
            mask = filetbl['loc'].isin(loclist)
            filetbl = filetbl[mask].reset_index(drop=True)

        return cls(filetable=filetbl, source='dinocsv')

    @property
    def filetable(self):
        return self._ftb

    def iteritems(self):
        """Iterate over all series and return gwseries object."""
        for idx,row in self._ftb.iterrows():

            if self._source == 'dinocsv':
                gw = GwSeries.from_dinogws(row['fpath'])

            if self._source == 'json':
                gw = GwSeries.from_json(row['fpath'])

            yield gw

    def to_json(self,dirpath):
        """Write all gwseries to json files.

        Parameters
        ----------
        dirpath : str
            Valid output directory for json files.

        Returns
        -------
        List of json objects.
        """
        jsonlist = []
        for gw in self.iteritems():
            jsonlist.append(gw.to_json(dirpath))
        return jsonlist

    def to_csv(self,dirpath):
        """Write all gwseries to json files.

        Parameters
        ----------
        dirpath : str
            Valid output directory for json files.

        Returns
        -------
        List of pandas series.
        """
        csvlist = []
        for gw in self.iteritems():
            csvlist.append(gw.to_csv(dirpath))
        return csvlist


    # Classmethods below are defined here to make sure input files have 
    # allready been created by code above.

    @classmethod
    def from_json(cls,srcdir,loclist=None):
        """Create GwFiles object from folder with json sourcefiles.
        
        Parameters
        ----------
        srcdir : str
            Path to directory with json GwSeries sourcefiles.
        loclist : list, optional
            List of strings with valid location names to restrict
            number of files read from srcdir.
        """

        if not os.path.isdir(srcdir):
            raise ValueError(f'Directory {srcdir} does not exist')

        # table of filenames
        filenames = [fname for fname in os.listdir(srcdir) if fname.endswith('.json')]
        filetbl = pd.DataFrame({"fname":filenames})

        # add columns to filetbl
        filetbl["fpath"]= filetbl["fname"].apply(lambda x:srcdir+x)
        filetbl.insert(0,"loc",filetbl["fname"].apply(lambda x:x.split('_')[0]))
        filetbl.insert(1,"fil",filetbl["fname"].apply(lambda x:x.split("_")[-1].lstrip("0")))
        filetbl.insert(0,"series",filetbl["loc"]+"_"+filetbl['fil'])

        if loclist is not None:
            mask = filetbl['loc'].isin(loclist)
            filetbl = filetbl[mask].reset_index(drop=True)

        return cls(filetable=filetbl, source='json')




"""
Module with GwCollection object holding a collection of groundwater 
head series.
"""

import warnings
from pandas import Series, DataFrame
import pandas as pd
from shapely.geometry import Point
import geopandas as gpd

from .read.gwfiles import GwFiles
from .read.waterweb import WaterWeb
from .read.hydromonitor import HydroMonitor

class GwCollection:
    """Collection of groundwater head series."""

    def __init__(self, gwcol):

        self._collection = gwcol
        self._stats = None
        self._xg = None
        self._ref = None

    def __len__(self):    
        return len(self._collection)

    def __repr__(self):
        return f'{self.__class__.__name__} (n={len(self)})'
    
    @classmethod
    def from_dinocsv(cls,filedir,loclist=None):
        """Create GwCollection object from folder with DinoLoket sourcefiles.
        
        Parameters
        ----------
        srcdir : str
            Path to directory with Dinoloket csv sourcefiles.
        loclist : list, optional
            List of strings with valid location names to restrict
            number of files read from srcdir.
        """
        gwcol = GwFiles.from_dinocsv(filedir,loclist=loclist)
        return cls(gwcol)

    @classmethod
    def from_json(cls,filedir,loclist=None):
        """Create GwCollection object from folder with json sourcefiles.
        
        Parameters
        ----------
        srcdir : str
            Path to directory with Dinoloket csv sourcefiles.
        loclist : list, optional
            List of strings with valid location names to restrict
            number of files read from srcdir.
        """
        gwcol = GwFiles.from_json(filedir,loclist=loclist)
        return cls(gwcol)

    @classmethod
    def from_waterweb(cls, filepath):
        """Create GwCollection object from WaterWeb csv export file.
        
        Parameters
        ----------
        srcdir : str
            Path to directory with Dinoloket csv sourcefiles.

        Returns
        -------
        GwCollection object
        ...
        """
        wwn = WaterWeb.from_csv(fpath=filepath)
        return cls(wwn)

    @classmethod
    def from_hydromonitor(cls, filepath):
        """Create GwCollection object from Menyanthes hydromonitor csv export file.
        
        Parameters
        ----------
        srcdir : str
            Path to directory with Dinoloket csv sourcefiles.

        Returns
        -------
        GwCollection object
        ...
        """
        hm = HydroMonitor(fpath=filepath)
        return cls(hm)


    def iteritems(self):
        """Iterate over all series in collecion and return gwseries 
        object."""
        for gw in self._collection.iteritems():
            yield gw


    def _calculate_series_stats(self,ref=None):
        """Return table with series statistics.

        Parameters
        ----------
        ref : {'datum','surface','mp'}, default 'datum'
            head reference level
        """
        srstats_list = []
        xg_list = []
        for gw in self.iteritems():

            if not gw.tubeprops().empty:
                desc = gw.describe(ref=ref,gxg=True)
                srstats_list.append(desc)
                xg_list.append(gw.xg(ref=ref,name=True))
            else:
                warnings.warn((f'{gw.name()} has no tubeproperties ' 
                    f' and will be ignored.'))

        self._stats = pd.concat(srstats_list,axis=1).T
        self._stats.index.name = 'series'
        self._xg = pd.concat(xg_list,axis=0)
        self._ref = ref


    def get_headstats(self,ref='datum'):
        """Return statistics of measured heads for each well filter."""
    
        # recalculate stats if nessesary
        if (self._stats is None) | (self._ref!=ref):
            self._calculate_series_stats(ref=ref)


        xcr = self._stats['xcr'].astype('float').values
        ycr = self._stats['ycr'].astype('float').values
        geometry = [Point(crd) for crd in zip(xcr,ycr)]
        points = gpd.GeoDataFrame(self._stats, geometry=geometry)
        points = points.set_crs('EPSG:28992')

        for colname in ['firstdate','lastdate',]:
            points[colname] = points[colname].apply(lambda x:x.strftime('%d-%m-%Y'))

        return points


    def get_xg(self,ref='datum'):
    
        # recalculate stats if nessesary
        if (self._stats is None) | (self._ref!=ref):
            self._calculate_series_stats(ref=ref)
        
        return self._xg


    def get_ecostats(self,ref='surface', units='days', step=5):
        """Return ecological most relevant statistics.

        Parameters
        ----------
        ref : {'datum','surface'}, default 'surface'
            Reference level for measurements.
        units : {'days','quantiles'}, default 'days'
            Unit of quantile boundary classes.
        step : float or int, default 5
            Quantile class division steps. For unit days an integer 
            between 0 and 366, for unit quantiles a fraction between 
            0 and 1.

        Returns
        -------
        pd.DataFrame
        ..."""

        ecostats = []
        for gw in self.iteritems():
            ecostats.append(gw.get_ecostats())
        return DataFrame(ecostats)

    def get_timestats(self, ref='datum', asgeo=True):

        statslist = []
        for gw in self.iteritems():

            # timestats table
            stats = gw.timestats(ref=ref)
            stats.name = gw.locname()
            stats = pd.DataFrame(stats).T
            locprops = gw.locprops()
            stats = pd.merge(stats,locprops,left_index=True,right_index=True, how='left')
            
            # order columns
            #firstcols = ['locname','filname']
            #colnames = firstcols + [col for col in stats.columns if col not in firstcols]

            statslist.append(stats.copy())
        stats = pd.concat(statslist)

        # create geodataframe
        if asgeo:
            xcr = stats['xcr'].astype('float').values
            ycr = stats['ycr'].astype('float').values
            geometry = [Point(crd) for crd in zip(xcr,ycr)]
            stats = gpd.GeoDataFrame(stats, geometry=geometry)
            stats = stats.set_crs('EPSG:28992')

        return stats




"""This module contains the HydroMonitor object for reading groundwater
head measurements from a HydroMonitor csv export file
"""

from collections import OrderedDict
import warnings
import os.path
import errno
import os

import matplotlib as mpl
import matplotlib.pyplot as plt
from pandas import Series, DataFrame
import pandas as pd
import numpy as np

from ..gwseries import GwSeries


class HydroMonitor:
    """Read and manage data from hydromonitor csv export file

    Methods
    -------
    get_series
        Return GwSeries object from HydroMonitor object.
    to_list
        Return time series data as list of GwSeries objects.
    to_json
        Save all data to json files.
    iterdata
        Return generator for iterating over heads data.
    idkeys
        Return column names that uniquely identify a series.
    delete_duplicate_data
        Remove duplicate data from data and return DataFrame.

    Examples
    --------
    Read hydromonitor csv export file:
    >>>hm = HydroMonitor(filepath=<path>)
    Convert to list of GwSeries objects:
    >>>mylist = hm.to_list()
    Iterate over all series and return GwSeries objects one at a time:
    >>>for i in range(len(hm)):
        gw = next(hm)
    Iterate over raw data and returnGwSeries objects: 
    >>>for (loc,fil),data in hm.iterdata():
        gw = hm.get_series(data=data,loc=loc,fil=fil)
    Save all series to json files in <filedir>:
    >>>hm.to_json(<filedir>)
    """

    CSVSEP = ";"
    METATAG = 'StartDateTime'
    DATATAG = 'DateTime'

    META_NUMCOLS = ['XCoordinate','YCoordinate','SurfaceLevel',
        'WellTopLevel','FilterTopLevel','FilterBottomLevel',
        'WellBottomLevel',]

    MAPPING_TUBEPROPS = OrderedDict([
        ('startdate','StartDateTime'),
        ('mplevel','WellTopLevel'),
        ('filtop','FilterTopLevel'),
        ('filbot','FilterBottomLevel'),
        ('surfacedate',None),
        ('surfacelevel','SurfaceLevel'),
        ])

    def __init__(self,fpath):
        """
        Parameters
        ----------
        fpath : str
            Valid filepath to hydromonitor csv file.
        """
        self.fpath = fpath

        # read header and line numbers from csv
        self.header, self._line_numbers, self.meta_colnames, self.data_colnames = self._readcsv()

        # extract metadata and data from file
        self.metadata, self.data = self._extract_contents()

        # remopve duplicate data
        self.data = self._delete_duplicate_data()

        # create generator
        self._srgen = self.data.groupby(self.idkeys).__iter__()
        self._itercount = 0


    def __repr__(self):
        return (f'{self.__class__.__name__}') # ({len(self.metadata)} filters)')


    def _readcsv(self):
        """ Read hydromonitor csv export file """

        try:
            textfile = open(self.fpath)
        except IOError:
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), self.fpath)
            textfile = None
        else:
            header, line_numbers, metacols, datacols = self._read_header(textfile)
            textfile.close()

        return header, line_numbers, metacols, datacols


    def _read_header(self,textfile):
        """ 
        read header and linenumbers from hydromonitor export file 
        returns
        -------
        header : pandas dataframe
            header items as pandas dataframe
        filelines : tuple
            line numbers
        """

        metatag_found = False
        datatag_found = False

        header = OrderedDict()
        for i,line in enumerate(textfile):

            line = line.rstrip()
            linelist = line.split(self.CSVSEP)

            # read header tags
            if line.startswith('Format Name'):
                header['format_name']=linelist[1]
            elif line.startswith('Format Version'):
                header['format_version']=linelist[1]
            elif line.startswith('Format Definition'):
                header['format_definition']=linelist[1]
            elif line.startswith('File Type'):
                header['file_type']=linelist[1]
            elif line.startswith('File Contents'):
                values = list(filter(None, linelist[1:]))
                header['file_contents']=values
            elif line.startswith('Object Type'):
                header['object_type']=linelist[1]
            elif line.startswith('Object Identification'):
                values = list(filter(None, linelist[1:]))
                header['object_identification']=values

            # read metadata line number and column names
            elif self.METATAG in line:
                metatag_found = True
                metafirst = i+2
                #metacols = [x.lower() for x in linelist]
                metacols = [x for x in linelist]

            # read data line number column names
            elif self.DATATAG in line:
                datatag_found = True
                datafirst = i+2
                metalast = i-2
                ##datacols = [x.lower() for x in linelist]
                datacols = [x for x in linelist if len(x)>0]
                break # avoid iterating over lines after metadata

        # warnings
        if not metatag_found:
            warnings.warn(f'Metadata header {self.METATAG} not found.')

        if not datatag_found:
            warnings.warn(f'Data header {self.DATATAG} not found.')

        # return variables
        header = Series(header)
        line_numbers = Series({'metafirst':metafirst,'metalast':metalast,
            'datafirst':datafirst},name='line_numbers')

        return header, line_numbers, metacols, datacols


    def _extract_contents(self):

        content_list = [x for x in self.header['file_contents']]
        if 'Metadata' in content_list:
            # this is where the work gets done
            metadata = self._read_metadata()
        else:
            metadata = None
            warnings.warn("Metadata not specified.")

        if 'Data' in content_list:
            if metadata is None:
                raise((f'Heads from {filepath} can not be read '
                      f'because metadata are not available.'))
            # this is where the work gets done
            data = self._read_data()
        else:
            data = None
            warnings.warn("Data not specified.")

        return metadata,data

    def _read_metadata(self):
        """ read metadata from hydromonitor csv export file """

        #Name;NITGCode;OLGACode;FilterNo;StartDateTime;XCoordinate;
        #YCoordinate;SurfaceLevel;WellTopLevel;FilterTopLevel;
        #FilterBottomLevel;WellBottomLevel;LoggerSerial;LoggerDepth;
        #Comment;CommentBy;Organization;Status;

        firstrow = self._line_numbers['metafirst'] 
        lastrow = self._line_numbers['metalast']
        nrows = (lastrow - firstrow + 1)

        # use pandas csv reader to read metadata
        meta = pd.read_csv(
            self.fpath,
            sep=self.CSVSEP,
            index_col=False,
            header=None,
            names=self.meta_colnames,
            skiprows=firstrow,
            nrows=nrows,
            dtype=str,
            encoding='latin-1',
            )

        # delete empty last column
        if '' in list(meta.columns):
            meta = meta.drop([''], axis=1)

        # convert datetime column
        meta['StartDateTime'] = pd.to_datetime(meta['StartDateTime'],
            format='%d-%m-%Y %H:%M',errors='coerce')

        for col in self.META_NUMCOLS:
            meta[col] = pd.to_numeric(meta[col],errors='coerce')

        return meta

    def _read_data(self):
        """ read data from hydromonitor csv export file """

        #read data
        colidx = list(range(len(self.data_colnames)))
        data = pd.read_csv(
            self.fpath,
            sep=self.CSVSEP,
            index_col=False,
            header=None,
            names=self.data_colnames,
            usecols = colidx,
            skiprows=self._line_numbers['datafirst'],
            #parse_dates=['datetime'], # don't, this takes a lot of time
            dtype=str,
            encoding='latin-1',
            )

        self._rawdata = data

        if 'LoggerHead' not in self.data_colnames:
        # when no loggerhead is available, menyanthes only exports
        # the column manualheads and the column loggerheads is simply missing
        # this happens when loggerdata without manual control measurments
        # are imported from another source; Menyanthes marks these 
        # measurements as manual heads.

            pos = data.columns.get_loc('DateTime')+1
            data.insert(loc=pos,column='LoggerHead',value=np.nan)

            msg = f'Missing data column LoggerHead added and filled with NaNs'
            warnings.warn(msg)

        # convert head columns to numeric
        for colname in ['LoggerHead','ManualHead']:
            if colname in self.data_colnames:
                data[colname] = data[colname].str.replace(',','.')
                data[colname] = pd.to_numeric(data[colname],errors='coerce')

        if 'ManualHead' not in self.data_colnames:
        # this is a variation on the previous missing loggerhead issue

            pos = len(data.columns)
            data.insert(loc=pos,column='ManualHead',value=np.nan)
            #dfdata['manualhead'] = np.nan

            msg = f'Missing data column LoggerHead added and filled with NaNs'
            warnings.warn(msg)


        # delete empty last column
        if '' in list(data.columns):
            data = data.drop([''], axis=1)

        # delete repeating headers deep down list of data as a result of
        # annoying bugs in menyanthes export module
        namecol = self.data_colnames[0]
        #data = data[data[namecol]!='NITGCode'].copy()
        #data = data[data[namecol]!='Name'].copy()
        data = data[data[namecol]!=namecol].copy()
        data = data[data[namecol]!='[String]'].copy()

        # parsing these dates is very time consuming
        data['DateTime'] = pd.to_datetime(data['DateTime'],
              dayfirst=True,format='%d-%m-%Y %H:%M',errors='coerce')


        return data

    @property
    def idkeys(self):
        """Return column names that give a unique identification of a 
        series """
        # bug in hydromonitor export? id is allways nitgcode, filterno
        # idkeys = ['nitgcode','filterno']
        if len(self.header['object_identification'])>2:
            message.warn('More than two object identification keys given')
        return [x for x in self.header['object_identification']]


    def _delete_duplicate_data(self):
        """Remove duplicate data from data and return pd.DataFrame
        Duplicates occur in groundwater head measurments in
        hydromonitor export when loggervalue and manual control 
        measurement have the same timestamp."""
        sortcols = self.idkeys + ['DateTime','ManualHead']
        data = self.data.sort_values(by=sortcols)
        dropcols = self.idkeys + ['DateTime']
        self.data_no_dups = data.drop_duplicates(subset=dropcols, 
                            keep='first').copy()
        return self.data_no_dups


    def get_series(self,loc=None,fil=None):
        """Return GwSeries object from HydroMonitor object
        Parameters
        ----------
        loc : str
            Well location name
        fil : str
            Tube name
        Returns
        -------
        GwSeries object
        """

        gws = GwSeries()
        
        # create DataFrame with HydroMonitor metadata for one series
        bool_loc = self.metadata[self.idkeys[0]]==loc
        bool_fil = self.metadata[self.idkeys[1]]==fil
        metadata = self.metadata[bool_loc & bool_fil]
        if metadata.empty:
            raise ValueError((f"Combination of loc='{loc}' and fil='{fil}' "
                f"not found in HydroMonitor metadata."))

        # Metadata can have a new row of metadata for each change.
        # Therefore, metaqdata can one or mulitple rows. For GwSeries
        # locprops items
        # all rows have the same value and metadata are simply taken 
        # from the first row.
        idx_firstrow = metadata.index[0]

        # create DataFrame with Hydromonitor measurements for one series
        bool_loc = self.data[self.idkeys[0]]==loc
        bool_fil = self.data[self.idkeys[1]]==fil
        data = self.data[bool_loc & bool_fil]

        # GwSeries tubeprops from HydroMonitor metadata
        for prop in GwSeries._tubeprops_names: #list(gws._tubeprops):
            metakey = self.MAPPING_TUBEPROPS[prop]
            if metakey is not None:
                gws._tubeprops[prop] = metadata[metakey].values
            if prop not in self.MAPPING_TUBEPROPS.keys():
                warnings.warn(f"Unknown property {prop} in {type(gws)}")

        # GwSeries locprops from HydroMonitor metadata
        for prop in GwSeries._locprops_names: ##list(gws._locprops.index):

            if prop=='locname':
                gws._locprops[prop] = metadata.at[idx_firstrow,self.idkeys[0]]
            
            if prop=='filname':
                gws._locprops[prop] = metadata.at[idx_firstrow,self.idkeys[1]]

            if prop=='alias':
                if 'NITGCode' in self.idkeys: 
                    alias_key = 'Name'
                else: 
                    alias_key = 'NITGCode'
                gws._locprops[prop] = metadata.at[idx_firstrow,alias_key]

            if prop=='xcr':
                gws._locprops[prop] = metadata.at[idx_firstrow,'XCoordinate']

            if prop=='ycr':
                gws._locprops[prop] = metadata.at[idx_firstrow,'YCoordinate']

            if prop=='height_datum':
                gws._locprops[prop] = 'mnap'

            if prop=='grid_reference':
                gws._locprops[prop] = 'rd'

            if prop not in GwSeries._locprops_names:
                warnings.warn(f"Unknown property {prop} in {type(gws)}")

        # set gwseries
        datetimes = data['DateTime'].values
        heads = np.where(
            np.isnan(data['LoggerHead'].values),
            data['ManualHead'].values,
            data['LoggerHead'].values
            )
        rec = {}
        for key in GwSeries._headprops_names:
            if key=='headdatetime':
                rec[key] = datetimes
            elif key=='headmp':
                rec[key] = heads
            else: # all other columns in GwSeries headprops
                rec[key] = np.full(len(heads),np.nan)
        gws._heads = DataFrame(rec)

        
        # remove rows with invalid datevalues and warn
        mask = pd.isna(gws._heads['headdatetime'])
        number_of_bad_dates = len(heads[mask])
        if number_of_bad_dates!=0:
            msg = f"Removed {number_of_bad_dates} rows with invalid date from {gws.name()}"
            warnings.warn(msg)
            heads = heads[~mask].copy()

        # convert heads in mnap to mref
        heads = gws._heads['headmp'].copy()
        dates = gws._heads['headdatetime'].copy()
        for idx, props in gws._tubeprops.iterrows():
            try:
                mask = dates >= props['startdate'] 
                heads = heads.mask(mask,props['mplevel']-heads)
            except TypeError as err:
                # probably mplevel is NaN
                warnings.warn((f'Warning: TypeError {err} on {gws.name()}'))
                heads = heads.mask(mask,np.nan-heads)
        gws._heads['headmp'] = heads
        return gws

    @property
    def series(self):
        """Return list of tuples wits series name and filter."""
        tbl = self.metadata[self.idkeys]
        srlist = []
        for idx2,row in tbl.iterrows():
            srtuple = (row[self.idkeys[0]],row[self.idkeys[1]])
            srlist.append(srtuple)
        return srlist
        

    @property
    def locations(self):
        """Return list of tuples with series name and filter grouped 
        by well locations."""
        locations = []        
        tbl = self.metadata[self.idkeys]
        for _,grp in tbl.groupby(by=self.idkeys[0]):
            srnames = []
            for _,row in grp.iterrows():
                srtuple = (row[self.idkeys[0]],row[self.idkeys[1]])
                srnames.append(srtuple)
            locations.append(srnames)       
        return locations

    def to_list(self):
        """ Return data from HydroMonitor as a list of GwSeries() 
            objects """

        srlist = []
        """
        heads = self._delete_duplicate_data()
        filgrp = heads.groupby(self.idkeys)
        for (location,filnr),data in filgrp:
            gws = self.get_series(loc=location,fil=filnr)
        """
        for gw in self.iteritems():
            srlist.append(gw)

        return srlist

    """
    # code below is superseded by iteritems method
    
    def __iter__(self):
        return self


    def __next__(self):

        if self._itercount >= len(self):
            raise StopIteration

        (loc,fil),data = next(self._srgen)
        gw = self.get_series(loc=loc,fil=fil)

        self._itercount+=1
        return gw

    def iterdata(self):
        #Return generator for iterating over heads data
        heads = self._delete_duplicate_data()
        return heads.groupby(self.idkeys).__iter__()
    """

    def __len__(self):
        heads = self._delete_duplicate_data()
        hymlen=0
        for srname,sr in heads.groupby(self.idkeys).__iter__():
            hymlen+=1
        return hymlen

    def to_json(self,filedir=None):

        for gw in self.iteritems():
            gw.to_json(filedir)


    def iteritems(self):
        """Iterate over all series and return gwseries object.
        
        Examples
        --------
        for gw in hm.iteritems():
            print(gw)
        """
        locs = self.metadata.loc[:,self.idkeys[0]].values
        fils = self.metadata.loc[:,self.idkeys[1]].values
        for loc, fil in zip(locs,fils):
            yield self.get_series(loc,fil)



"""Descriptive statistics of list of groundwater head series"""


import os
import warnings
from pandas import Series, DataFrame
import pandas as pd
import numpy as np

from ..gwlist import GwList

def gwliststats(srcdir=None,ref=None,gxg=False):
    """Return table of decriptive statistics for multiple heads series

    Parameters
    ----------
    srcdir : str
        directory name with groundwater head source files
    ref : {'datum','surface','mp'}, default 'datum'
        head reference level
    srctype : str, optional
        sourcefile type (will be inferred if not given)
    locs : list, optional
        list of location names
    gxg : bool, default False
        include GxG descriptive statistics

    Return
    ------
    pd.DataFrame

    """

    ds = GwListStats(srcdir)
    tb = ds.srstats(gxg=gxg)

    return tb


def gwlocstats(srstats):
    """Return table of descriptive statistics for groundwater well 
    locations summarised from table of heads series statistics

    Parameters
    ----------
    srstats : pd.DataFrame
        table with series stats (as returned by  
        GwListTimeStats.timestatstable()

    Returns
    -------
    pd.DataFrame

    """
    gws = GwListStats(srstats=srstats)
    return gws.locstats()


class GwListStats:
    """Return table of decriptive statistics for list of heads series

    Methods
    -------
    srstats(ref='datum',gxg=False)
        Return series statistics
    xg()
        Return xg3 statistics for all series
    locstats():
        Return location statistics
    save(fdir,ref='datum',gxg=False,prefix=None,suffix=None)
        Save all tables to directory

    Examples
    --------
    srstats = gst.srstats(ref='datum',gxg=True)
    locstats = gst.locstats()
    xg = gst.xg()
    gst.save(<valid directory>,ref='datum',gxg=True)
       
       
    """
        
        
    def __init__(self,srcdir=None,locs=None,gwlist=None,srstats=None):
        """Create GwListStats object

        Parameters
        ----------
        srcdir : str
            directory name with groundwater head source files
        locs : bool, default False
            aggegate results to locations
        gwlist : aq.GwList object, optional
            list of gwseries objects
        srstats : pd.DataFrame
            table with series statistics as returned by stats() method

        Examples
        --------
        gls = aq.GwListStats(srcdir=<directory with dinofiles>)
        gls = aq.GwListStats(gwlist=<GwList object)

        """

        self._srcdir = srcdir
        self._locs = locs
        self._gwlist = gwlist
        self._srstats = srstats

        #if (self._gwlist is None) and (self._srcdir is None):
        #    raise ValueError(
        #        (f'For creating a GwListStats object, '
        #         f'either a GwList object or a valid source file '
        #         f'directory must be given.'))

        if self._srcdir is not None:
            if not os.path.isdir(self._srcdir):
                raise ValueError((f'{self._srcdir} is not a valid '
                    f'directory name.'))

        

    def __repr__(self):
        if not self._gwlist:
            mylen = 0
        else:
            mylen = len(self._gwlist)
        return (f'{self.__class__.__name__}(n={mylen})')


    def _create_list(self):
        """Create aq.GwList object"""
        self._gwlist = GwList(srcdir=self._srcdir,loclist=self._locs)


    def srstats(self,ref='datum',gxg=False):
        """Return series statistics

        Parameters
        ----------
        ref : {'datum','surface','mp'}, default 'datum'
            head reference level
        gxg : bool, default False
            include GxG descriptive statistics
        """

        if self._gwlist is None:
            self._create_list()

        srstats_list = []
        xg_list = []
        for i,gw in enumerate(self._gwlist):

            if not gw._tubeprops.empty:
                desc = gw.describe(ref=ref,gxg=gxg)
                srstats_list.append(desc)
                xg_list.append(gw.xg(ref=ref,name=True))
            else:
                warnings.warn((f'{gw.name()} has no tubeproperties ' 
                    f' and will be ignored.'))

        self._srstats = pd.concat(srstats_list,axis=1).T
        self._srstats.index.name = 'series'

        self._xg = pd.concat(xg_list,axis=0)

        return self._srstats


    def xg(self):
        """Return xg3 statistics for all series"""

        if not hasattr(self,'_xg'):
            self.srstats()

        return self._xg


    def locstats(self):
        """Return location statistics"""

        def get_maxfrq(sr):
            """Return maximum observation frequency for a series"""
            frqs = ['daily','14days','month','seldom','never']

            if sr.empty:
                return None

            for freq in frqs:
                if np.any(sr==freq):
                    return freq


        def get_maxdif(sr):
            """Calculate difference between maximum and minimum value
            in a series"""
            return np.round((sr.max()-sr.min())*100)


        aggdict = {
           'locname':'first',
           'filname':'size',
           'alias':'first',
           'surfacelevel':'first',
           'filbot':'min',
           'xcr':'first',
           'ycr':'first',
           'firstdate':'min',
           'lastdate':'max',
           'minyear':'min',
           'maxyear':'max',
           'nyears':'max',
           'yearspan':'max',
           'maxfrq':[get_maxfrq],
           'mean':[get_maxdif],
          }


        if self._srstats is None:
            self._srstats = self.srstats()


        """
        missing_cols = [x for x in srstats.keys() 
            if x not in aggdict.keys()]
        if len(missing_cols)!=0:
            msg = f'Missing columns in aggdict :{missing_cols}'
            warnings.warn(msg)
        """

        tbloc = self._srstats.groupby(by=['locname']).agg(aggdict)
        tbloc.columns = tbloc.columns.get_level_values(0)
        tbloc = tbloc.drop('locname',axis=1)

        coldict = {'filname':'nfil','mean':'meandifcm'}
        tbloc = tbloc.rename(columns=coldict)

        return tbloc


    def save(self,fdir,ref='datum',gxg=False,prefix=None,suffix=None):
        """Save all tables to directory

        Parameters
        ----------
        fdir : str
            directory to save tables to 
        ref : {'datum', surface'}, default 'datum'
            reference level for heads
        prefix : str, optional
            prefix for filenames
        gxg : bool, default False
            include gxg statistics
        """
        if prefix is None:
            prefix = ''
 
        if not isinstance(prefix,str):
            warnings.warn((f'parameter prefix is of type {type(prefix)} ', 
                f'not type str. Prefix will be ignored.'))
            prefix = ''

        if suffix is None:
            suffix = ''
 
        if not isinstance(suffix,str):
            warnings.warn((f'parameter suffix is of type {type(suffix)} ', 
                f'not type str. Suffix will be ignored.'))
            suffix = ''

        if self._gwlist is None:
            self._create_list()

        if not self._gwlist.is_callable():
            self._create_list()

        # series statistics
        srstats = self.srstats(ref=ref,gxg=gxg)
        outfilepath = f'{fdir}{prefix}srstats{suffix}'
        srstats.to_csv(f'{outfilepath}.csv',index=True)
        srstats.to_excel(f'{outfilepath}.xlsx',index=True,merge_cells=False)

        # location statistics
        locstats = self.locstats()
        outfilepath = f'{fdir}{prefix}locstats{suffix}'
        locstats.to_csv(f'{outfilepath}.csv',index=True)
        locstats.to_excel(f'{outfilepath}.xlsx',index=True,merge_cells=False)

        # xg statistcs for each year
        if gxg==True:
            xg = self.xg()
            outfilepath = f'{fdir}{prefix}xg{suffix}'
            xg.to_csv(f'{outfilepath}.csv',index=True)
            xg.to_excel(f'{outfilepath}.xlsx',index=True,merge_cells=False)





import pytest
import pandas as pd
from geopandas import GeoDataFrame
from acequia import measurement_types
from acequia import WaterWeb
from acequia import GwSeries

fpath = r'.\data\waterweb\grolloer koelanden.csv'
networkname = 'grolloer koelanden'
wwdir = r'.\data\waterweb\\'

# test class WaterWeb

@pytest.fixture
def wwn():
    return WaterWeb.from_csv(fpath=fpath,network=networkname)

def test_init():
    wwninit = WaterWeb(fpath=fpath,network=networkname)
    assert len(wwninit.srnames)==0

def test_from_csv(wwn):
    assert len(wwn.srnames)!=0

def test_get_type(wwn):
    srtype = wwn.get_type(wwn.srnames[0])
    assert isinstance(srtype,str)
    assert srtype in wwn.MEASUREMENT_TYPES

def test_measurement_types(wwn):
    assert isinstance(wwn.measurement_types,pd.Series)

def test_get_locname(wwn):
    assert isinstance(wwn.get_locname(wwn.srnames[0]),str)

def test_get_filname(wwn):
    assert isinstance(wwn.get_filname(wwn.srnames[0]),str)

def test_networkname_setter(wwn):
    original_name = wwn.networkname
    wwn.networkname = 'different name'
    assert isinstance(wwn.networkname,str)
    assert wwn.networkname!=original_name

def test_get_locprops(wwn):
    locprops = wwn.get_locprops(wwn.srnames[0])
    assert isinstance(locprops,pd.Series)
    assert locprops.empty is False

def test_get_tubeprops(wwn):
    tubeprops = wwn.get_tubeprops(wwn.srnames[0])
    assert isinstance(tubeprops,pd.DataFrame)
    assert tubeprops.empty is False

def test_get_levels(wwn):
    levels = wwn.get_levels(wwn.srnames[0])
    assert isinstance(levels,pd.Series)
    assert levels.empty is False

def test_get_gwseries(wwn):
    srname = wwn.srnames[0]
    gw = wwn.get_gwseries(srname)
    assert isinstance(gw.heads(),pd.Series)
    assert gw.heads().empty is False

def test_locations(wwn):
    locs = wwn.locations
    assert isinstance(locs,GeoDataFrame)
    assert not locs.empty

def test_to_kml(wwn):
    outpath = f'.\\output\\kml\\{wwn.networkname}.kml'
    wwn.to_kml(outpath)

def test_iteritems(wwn):
    for gw in wwn.iteritems():
        assert isinstance(gw,GwSeries)

# test custom functions in module waterwebtools

def test_measurement_types():
    tbl = measurement_types(wwdir)
    assert isinstance(tbl,pd.DataFrame)
    assert tbl.empty is False




import pytest
from pandas import DataFrame
from geopandas import GeoDataFrame
from acequia import KnmiData


def test_data_wtr_stns():
    knmi = KnmiData()
    gdf = knmi.wtr_stns
    assert isinstance(gdf,GeoDataFrame)
    assert not gdf.empty

def test_data_prc_stns():
    knmi = KnmiData()
    gdf = knmi.prc_stns
    assert isinstance(gdf,GeoDataFrame)
    assert not gdf.empty





import numpy as np
from pandas import Series, DataFrame
from pandas.api.types import is_int64_dtype
import pandas as pd


def hydroyear(sr):
    """Return hydrological year for each date in timeseries

    Note
    ----
    A hydrological year strts on april 1st and ends on march 30th 
    """
    return np.where(sr.index.month<4,sr.index.year-1,sr.index.year)


def season(sr):
    """Return season for each date in timeseries"""
    cond1 = sr.index.month>3
    cond2 = sr.index.month<10
    return np.where(cond1&cond2,"summer","winter")


def index1428(minyear=None,maxyear=None,days=[14,28]):
    """ Return datetimeindex with only daynumber in days

    Parameters
    ----------
    minyear : number
        first year in DatetimeIndex
    maxyear : number
        first year in DatetimeIndex
    days : list, default [14,28]
        days in DatetimeIndex

    Returns
    -------
    pd.DatetimeIndex

    """

    strdates = [str(year)+"-"+str(month)+"-"+str(day) #+' 12:00' 
                for year in list(range(minyear,maxyear+1)) 
                for month in list(range(1,13)) for day in days]

    dates = [pd.Timestamp(x) for x in strdates]

    return pd.DatetimeIndex(dates)


def ts1428(sr,maxlag=0,remove_nans=True, days=[14,28]):
    """ Return timeseries of measurements on 14th and 28th of each 
    month

    Parameters
    ----------
    sr : pd.Series
        timeseries
    maxlag : integer, default 0
        maximum number of days a measurement is allowed to deviate
        from the 14th or 28th
    remove_nans : boolean, default True
        remove values with nans before first valid value
    days : list, default [14,28]
        days in DatetimeIndex

    Returns
    -------
    sr : pandas time Series

    Notes
    -----
    When maxlag=0 only measurements on the 14th and 28th are 
    selected. When maxlag is 1 or higher, missing values on the 
    14th and 28th are filled in with values from the nearest date,
    with a maximum deviation of maxlag days.

    """

    # create empty timeseries with 1428 index
    minyear = sr.index.min().year
    maxyear = sr.index.max().year
    idx1428 = index1428(minyear=minyear,maxyear=maxyear,days=days)
    ts1428 = pd.Series(data=np.nan, index=idx1428)

    # add values to timeseries ts1428
    for date in ts1428.index:

        daydeltas = sr.index - date # type(daydeltas) = TimedeltaIndex
        mindelta = np.amin(np.abs(daydeltas))
        sr_nearest = sr[np.abs(daydeltas) == mindelta]

        maxdelta = pd.to_timedelta(f'{maxlag} days')
        if (mindelta <= maxdelta) & (len(sr_nearest)!=0):
            ts1428[date] = sr_nearest.iloc[0]

    if remove_nans==True:
        ts1428 = ts1428[ts1428.first_valid_index():ts1428.last_valid_index()]

    return ts1428


def measfrqclass(n):
    """Return measurement frequency class given number of yearly 
    measurements n"""

    if n>27: 
        return "daily"
    elif n>12: 
        return "14days"
    elif n>9: 
        return "month"
    elif n>0:
        return "seldom"
    else: 
        return "never"


def measfrq(ts):
    """Return estimated measurement frequency for each year in a time 
    series"""

    yearfrq = ts.groupby(ts.index.year).count()
    yearfrq.index.name = 'year'
    return yearfrq.apply(measfrqclass)


def maxfrq(sr):
    """Return maximum of estimated yearly measurement frequencies in
    a time series.

    Input can be pd.Series with pd.DatetimeIndex or pd.Int64Index or
    a list or numpy array with measurement frequencies"""

    frqs = ['daily','14days','month','seldom','never']

    if isinstance(sr,pd.Series):

        if isinstance(sr.index,pd.DatetimeIndex):
            sr = measfrq(sr)

        #if isinstance(sr.index,pd.Int64Index):
        if is_int64_dtype(sr.index.dtype):

            if pd.to_numeric(sr, errors='coerce').notnull().all():
                sr = sr.apply(measfrqclass).values

            for freq in frqs:
                if np.any(sr==freq): 
                    return freq


    if isinstance(sr,np.ndarray) or isinstance(sr,list):

        if all(pd.notnull(pd.to_numeric(sr,errors='coerce'))):
            sr = [measfrqclass(x) for x in sr]

        for freq in frqs:
            if any([x==freq for x in sr]):
                return freq

    """
        ts = np.array(ts)
        allint = all([x.dtype=='int32' for x in ts])
        allfloat = all([x.dtype=='float64' for x in ts])

        # ts is a np.ndarray or list of numbers
        if allint or allfloat:
            ts = [measfrqclass(x) for x in ts]

    for freq in frqs:
        if np.any(ts==freq): 
            return freq
    """






import pytest
##import collections
##import numpy as np
from pandas import DataFrame
import pandas as pd
from geopandas import GeoDataFrame


from acequia import GwCollection, GwFiles, WaterWeb, GwSeries, HydroMonitor
import acequia as aq

dinodir = '.\\data\\dinogws_small\\'
jsondir = '.\\output\\json\\'
csvdir = '.\\output\\csv\\'
wwfile = '.\\data\waterweb\\grolloer koelanden.csv'
hmfile = r'.\\data\\hymon\\hydromonitor_testdata.csv'

@pytest.fixture
def gwc():
    # test_from_dinocsv()
    gwc = GwCollection.from_dinocsv(dinodir)
    assert isinstance(gwc, GwCollection)
    assert isinstance(gwc._collection, GwFiles)
    return gwc

@pytest.fixture
def gwc_json():
    # test_from_dinocsv()
    gwc = GwCollection.from_json(jsondir)
    assert isinstance(gwc, GwCollection)
    assert isinstance(gwc._collection, GwFiles)
    return gwc

@pytest.fixture
def gwc_ww():
    # test_from_waterweb()
    gwc = GwCollection.from_waterweb(wwfile)
    assert isinstance(gwc, GwCollection)
    assert isinstance(gwc._collection, WaterWeb)
    return gwc

@pytest.fixture
def gwc_hm():
    # test_from_hydromonitor()
    gwc = GwCollection.from_hydromonitor(hmfile)
    assert isinstance(gwc, GwCollection)
    assert isinstance(gwc._collection, HydroMonitor)
    return gwc


def test_len(gwc):
    assert len(gwc)!=0

def test_repr(gwc):
    assert isinstance(repr(gwc),str)

def test_ecostats(gwc):
    df = gwc.get_ecostats(ref='datum', units='days', step=5)
    assert isinstance(df, DataFrame)
    assert not df.empty

def test_headstats(gwc): 
    df = gwc.get_headstats(ref='datum')
    assert isinstance(df, DataFrame)
    assert not df.empty

def test_get_xg(gwc):
    df = gwc.get_xg(ref='datum')
    assert isinstance(df, DataFrame)
    assert not df.empty

def test_timestats(gwc):
    gdf = gwc.get_timestats(ref='datum', asgeo=True)
    assert isinstance(gdf,GeoDataFrame)
    assert not gdf.empty

def test_iteritems(gwc, gwc_json, gwc_ww, gwc_hm):
    for gw in gwc.iteritems():
        assert isinstance(gw, GwSeries)

    for gw in gwc_json.iteritems():
        assert isinstance(gw, GwSeries)

    for gw in gwc_ww.iteritems():
        assert isinstance(gw, GwSeries)

    for gw in gwc_hm.iteritems():
        assert isinstance(gw, GwSeries)




from pathlib import Path
import warnings
import numpy as np
import pandas as pd
from pandas import DataFrame, Series


class KnmiWeather:
    """Read weather data from knmi file

    Parameters
    ----------
    fpath : str
        path to knmi txt file with weather data

    Examples
    --------
    wtr = KnmiWeather.from_file(<filepath>)
    evp = wtr.timeseries('evp')
    name = wtr.name()
    """
    NAMESTR = ''.join([
        'STN,YYYYMMDD,DDVEC,FHVEC,   FG,  FHX, FHXH,  FHN, FHNH,',
        '  FXX, FXXH,   TG,   TN,  TNH,   TX,  TXH, T10N,T10NH,   SQ,',
        '   SP,    Q,   DR,   RH,  RHX, RHXH,   PG,   PX,  PXH,   PN,',
        '  PNH,  VVN, VVNH,  VVX, VVXH,   NG,   UG,   UX,  UXH,   UN,',
        '  UNH, EV24'])
    COLNAMES = NAMESTR.replace(' ','').split(',')
    KEEPCOLS = 'YYYYMMDD,RH,EV24'.split(',')
    VARNAMES = ['prc','evp','rch']
    SKIPROWS = 52

    def __init__(self,filepath=None):
        """Read Knmi Weather csv file.
        
        Parameters
        ----------
        filepath : str
            Valid filepath to csv file with weather data.
        """
        self.fpath = filepath
        self.rawdata = self._read_data(filepath)
        self.meta = self._read_meta(filepath)
        self.data = self._clean_rawdata(self.rawdata)
        self.stn = int(self.rawdata.loc[0,'STN'])

    def __repr__(self):
        return (f'{self.__class__.__name__} (n={len(self.data)})')

    def _read_data(self,filepath):

        # read data with pandas
        rawdata = pd.read_csv(filepath,sep=',',
            skiprows=self.SKIPROWS,
            names=self.COLNAMES,dtype='str')

        # replace empty strings with NaN
        rawdata = rawdata.apply(
            lambda x: x.str.strip()).replace('', np.nan)

        return rawdata

    def _read_meta(self,filepath):

        # read metadata from header
        with open(filepath, 'r') as fp:
            line_numbers = list(range(10,50))
            lines = []
            for i, line in enumerate(fp):
                if i in line_numbers:
                    line = line.strip()
                    desc = line[11:].strip()
                    rec = {
                        'variabele':line[0:10].strip(),
                        'omschrijving':desc.split('/')[0].strip(),
                        'description':desc.split('/')[1].strip(),
                        }
                    lines.append(rec)
                elif i > 49:
                    break
        desc = DataFrame(lines)
        return desc

    def _clean_rawdata(self,rawdata):

        data = rawdata[self.KEEPCOLS].copy()
        for colname in list(data):

            if colname=='YYYYMMDD':

                # create datetimeindex from string column
                data[colname] = pd.to_datetime(data[colname], #,format='%Y%m%d')
                    infer_datetime_format=True)
                data = data.set_index(
                    colname,verify_integrity=True)

                # make sure all dates are in index
                firstdate = data.index[0]
                lastdate = data.index[-1]
                idx = pd.date_range(start=firstdate, end=lastdate,
                    freq='D')
                data = data.reindex(idx)
                data.index.name='date'

            if colname=='RH':
                # RH is precipitation in 0.1 mm/day
                # RH = -1 means RH < 0.05 mm/day
                data[colname] = data[colname].replace('-1','0.5')

            if colname in ['RH','EV24']:
                data[colname] = data[colname].astype(float)/10.

        # drop nans and reindex
        data = data.dropna(how='all')
        newindex = pd.date_range(data.index.min(),data.index.max())
        data = data.reindex(newindex)

        return data

    def timeseries(self,var='prc'):
        """Return timeseries with data

        Parameters
        ----------
        var : {'prc','evp','rch'), default 'prc'
            variable to return
            
        Returns
        -------
        pd.Series
        """
        if var not in self.VARNAMES:
            msg = [f'{var} is not a valid variable name.',
                f'parameter varname must be in {self.VARNAMES}',
                f'by default rain data are returned']
            warnings.warn((f'{var} is not a valid variable name. ',
                f'parameter varname must be in {self.VARNAMES}'
                f'by default rain data are returned.'))
            var = 'prc'

        if var == 'prc':
            sr = self.data['RH']
            sr.name = 'prc'
        if var == 'evp':
            sr = self.data['EV24']
            sr.name = 'evp'
        if var == 'rch':
            sr = self.data['RH']-self.data['EV24']
            sr.name = 'rch'

        first = sr.first_valid_index()
        last = sr.sort_index(ascending=False).first_valid_index()
        return sr[first:last]

    @property
    def units(self):
        """Return table with definitions and units of variables"""
        tbl = pd.DataFrame({
            'variable' : ['prc','evp','rch'],
            'datacol' : ['RH','EV24','calculated'],
            'unit' : ['mm/day','mm/day','mm/day'], 
            })
        tbl = tbl.set_index('variable')
        return tbl

    @property
    def prc(self):
        """Return precipitation time series."""
        return self.timeseries(var='prc')

    @property
    def evp(self):
        """Return evaporation time series."""
        return self.timeseries(var='evp')

    @property
    def recharge(self):
        """Return recharge time series."""
        return self.timeseries(var='rch')





import pathlib
import collections
import datetime as dt
import numpy as np
from pandas import Series,DataFrame
import pandas as pd


class DinoSurfaceLevel:
    """Read TNO Dinoloket csv file with surface water level measurements """

    SEP = ','
    QUOTE = '"'
    SINGLEQUOTE = '\''
    SPACE = ' '

    HEADERNAMES = ['nitgcode','alias','xcr','ycr','startdate','enddate',
        'dummy']

    HEADERTYPES = {'nitgcode':str,'alias':str,'xcr':np.float64, 
        'ycr':np.float64,'startdate':str,'enddate':str,'dummy':str}

    DATANAMES = ['nitgcode','date','level','remark','dummy']

    DATATYPES = {'nitgcode':str,'date':str,'level':np.float64,
        'remark':str,'dummy':str}


    def __repr__(self):
        ##return (f'{self.__class__.__name__} (n={len(self._data)})')
        return (f'{self.name()} (n={len(self._data)})')


    def __init__(self,fpath=None):

        self._fpath = fpath

        if self._fpath is None:
            raise TypeError(f'Parameter fpath must be given when '
                f'calling {self.__class__.__name__}()')

        ispath = isinstance(fpath,pathlib.Path)
        isstr = isinstance(self._fpath,str)
        if not ispath and not isstr:
            raise TypeError(f'Parameter fpath must be of type filepath'
                f'or type str, not type {type(fpath)}')

        if not pathlib.Path(self._fpath).exists():
            raise ValueError(f'Filepath {self._fpath} not found')


        self._flines = self._readfile()
        self._meta = self._parselines()
        self._header = self._read_header()
        self._data = self._read_data()


    def _readfile(self):
        """Return DINO csv filelines as list of strings"""

        # read csv file to list of strings
        try:
            self._file = open(self._fpath,'r')
        except:  
            self._flines = []
            raise
        else:
            self._flines = self._file.readlines()
            self._file.close()

        return self._flines


    def _parselines(self):
        """Return dino csv metadata as table and find linenumbers with
        start and end of header and data"""

        # parse header data and find linenumbers
        meta = collections.OrderedDict()

        for i,line in enumerate(self._flines):

            ##line = line.rstrip() #remove trailing newline
            elms = self._flines[i].rstrip().split(',')

            if line.startswith('Titel'):
                pass

            if line.startswith('Gebruikersnaam'):
                pass

            if line.startswith('Periode aangevraagd'):

                dtm = dt.datetime.strptime(elms[1],'%d-%m-%Y')
                meta['qrstart'] = dtm

                dtm = dt.datetime.strptime(elms[3],'%d-%m-%Y')
                meta['qrend'] = dtm

            if line.startswith('Gegevens beschikbaar'):

                dtm = dt.datetime.strptime(elms[1],'%d-%m-%Y')
                meta['datastart'] = dtm

                dtm = dt.datetime.strptime(elms[3],'%d-%m-%Y')
                meta['dataend'] = dtm

            if line.startswith('Datum'):

                dtm = dt.datetime.strptime(elms[1],'%d-%m-%Y')
                meta['request'] = dtm

            if line.startswith('Referentie'):

                meta['reference'] = elms[1]

            if line.startswith(f'Locatie,Externe aanduiding,X-coordinaat,'
                f'Y-coordinaat, Startdatum, Einddatum'):

                self._rowstart_header = i+1

            if line.startswith(f'Locatie,Peildatum,Stand (cm t.o.v. NAP)'
                f',Bijzonderheid'):

                self._rowstart_data = i+2 #there is an extra whiteline...

        # count number of header rows
        self._rowcount_header=0
        for i in range(self._rowstart_header,self._rowstart_data):
            if len(self._flines[i].rstrip())==0:
                break
            self._rowcount_header+=1
 
        return Series(meta)


    def _read_header(self):
        """Return header lines as table"""

        self._header = pd.read_csv(self._fpath, sep=self.SEP, header=0,
            names=self.HEADERNAMES, skiprows=self._rowstart_header-2,
            nrows=self._rowcount_header,dtype=self.HEADERTYPES,
            on_bad_lines='warn')

        for colname in ['startdate','enddate']:
            self._header[colname] = pd.to_datetime(self._header[colname], 
                format='%d-%m-%Y', errors='coerce')

        return self._header


    def _read_data(self):
        """Return water level measurement data as table"""

        self._data = []
        for i,line in enumerate(self._flines):

            if i<self._rowstart_data:
                continue

            linedict = collections.OrderedDict()
            elms = self._flines[i].rstrip().split(self.SEP)

            # 'nitgcode','date','level', 'remarks','dummy'
            # process first 3 line elements
            for j in [0,1,2]:
                linedict[self.DATANAMES[j]] = elms[j]

            # proces remaining line elements as remarks
            stripchars = ''.join([self.SEP,self.SPACE,self.QUOTE,
                self.SINGLEQUOTE])
            remains = [x.strip(stripchars) for x in elms[3:]]
            remarks = self.SEP.join(remains).strip(stripchars)
            remarks = remarks.replace(self.SEP,'.')

            linedict[self.DATANAMES[3]] = remarks

            self._data.append(linedict)

        self._data = pd.DataFrame.from_dict(self._data)

        self._data['date'] = pd.to_datetime(self._data['date'], 
            format='%d-%m-%Y', errors='coerce')

        self._data['level'] = pd.to_numeric(self._data['level'],
            errors='coerce',downcast='integer')

        self._data['level'] = self._data['level']/100. #cm_to_m

        return self._data


    def locprops(self):
        """Return surface level gauge location properties"""
        colnames = ['nitgcode','alias','xcr','ycr']
        locp = self._header.loc[0,colnames]
        locp['reference'] = self._meta['reference']
        locp['grid_reference'] = 'RD'
        locp.name = 'locprops'
        return locp


    def levels(self):
        """Return surface water levels"""
        colnames = ['level'] #,'remark']
        levels = self._data.set_index('date')
        levels = levels[colnames]
        return levels


    def remarks(self):
        """Return table with dates and remarks"""
        mask = self._data['remark'].str.len()>0
        table = self._data[mask][['date','remark']]
        table = table.set_index('date',drop=True)
        return table


    def metadata(self):
        """Return metadata from header of dinocsv file"""
        return self._meta


    def name(self):
        """Return name of gauge"""
        return self._data.at[0,'nitgcode']




#-------------------------------------------------------------------------------
# Purpose:     plot a single pandas time serie to file
# Author:      Thomas de Meij
# Created:     29-03-2013 updated 3-6-2014
# Licence:     Python 2.7 with pandas and matplotlib
#-------------------------------------------------------------------------------
import warnings
from math import ceil as ceil, floor as floor
import os
import os.path
import datetime as dt


from pandas import DataFrame, Series
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.ticker import NullFormatter
from matplotlib.dates import YearLocator, MonthLocator, DateFormatter
import matplotlib.dates as mdates
import numpy as np
from .. import gwseries #import GwSeries directly gives circular import
#inch = 25.4 mm



class PlotHeads:
    """Class for plotting customized graph of groundwater head series

    """
    mm = 1/25.4

    # define line colors
    clrs = {
        'skytones': 
            ['#172458','#354898','#607cbd','#95b3d7','c','k','g'], # by rehenry
        'maximilian': 
            ['#d3d3d3','#e9e9e9','#a6bbca','#86a0b3','#65899d'], # by brynnreacjlocal
        'magazine2' : 
            ['#153f79','#79153f','#3f7915','#794f15','#1e1e1e'],
        'rainbow' : 
            ['#2347c5','#b41515','#1c9018','#ecee19','#ce11f2'],
        'easterpastel':
            ['#d666c6','#6e64db','#7ad89d','#67a3db','#dfe755'],
        'rainbow-adjusted':
            ['#2767fd','#d92950','#50d929','#ecee19','#ce11f2'],
        }

    font1 = {
        'family' : 'serif',
        'color'  : 'darkred',
        'weight' : 'normal',
        'size'   : 8,
        }


    def __init__(self,ts=[],ref="datum",lbs=None,mps=None,
        title=None,xlabel=None,ylabel=None,xlim=None,
        ylim=None,colors=None,plotargs=None,plot=True):
        """ Plot list of groundwater head series in one graph

        Parameters
        ----------
        ts : list, Series
            list of pd.Series with pd.DateTimeIndex
        ref : {'datum','surface','mp'}, default 'datum'
            reference level for groundwater heads
        lbs : list of strings, optional
            labels for heads series
        mps : pd.Series
            Timeseries with change of measurement reference point level
        title : string, optional
            title for graph
        xlabel : string, optional
            graph xaxis label
        ylabel :  string, optional
            graph yaxis label
        xlim : tuple or list of TimeStamp,int or str, optional
            min and max of dates on xaxis
        ylim : tuple or list, optional
            min and max of heads on yaxis
        colors : list,optional
            valid pyplot colors for drawing lines
        plotargs : list of dicts, optional
            dict of pyplot plotting parameters for each series
        plot : boolean, default True
            plot heads immediately (if Flase, call PlotHeads.plotheads()

        Notes
        -----
        xlim can have several forms:
        [2009,2012] plots period 1-1-2009:31-12-2012
        alternative: ['01-01-2009','31-12-2012]
        alternative: [<pd.Timestamp>,<pd.Timestamp>]

        Examples
        --------
        gws  = aq.GwSeries.from_json(<json sourcefile>)
        sr   = gw.heads(ref='datum')
        plot = aq.PlotHeads(ts=[sr],ylim=[19.5,21.5])
        plot.save(<filename.jpg>,dpi=300.0)
             
        """

        self.reflist = ['datum','surface','mp']
        if ref not in self.reflist:
            msg = f'Reference level must be in {self.reflist} not {ref}'
            raise TypeError(msg)
        self.ref = ref

        self.ts = ts
        # make sure self.ts is a list of pandas series with heads
        if isinstance(ts,Series):
            self.ts = [ts]
        if isinstance(ts,gwseries.GwSeries):
            self.ts = [ts]
        if all(isinstance(gw, gwseries.GwSeries) for gw in ts):
            self.ts = [gw.heads(ref=self.ref) for gw in ts]
        
        if self.ref=='surface':
            self.ts = [ts*100 for ts in self.ts]

        if not all(isinstance(sr, pd.Series) for sr in self.ts):
            raise TypeError((f'Wrong input of timeseries. Expected a '
                    f'list of Pandas Series with DateTime indexes.'))

        if not all(isinstance(sr.index, pd.DatetimeIndex) for sr in self.ts):
            raise TypeError(f'Index of series must be type DateTimeIndex')

        if (not isinstance(mps, pd.Series)) and (mps is not None):
            raise TypeError((f'Time series {mps} with referennce '
                f'changes must be pd.Series'))

        if isinstance(colors,list):
            self.clr = colors*10
        else:
            self.clr = self.clrs['rainbow-adjusted']*10 # avoid out of bounds


        self._validate_series()

        self.xlim = xlim
        self.ylim = ylim
        self.mps = mps
        self.title = title
        self.ylabel = ylabel
        self.xlabel = xlabel

        if lbs is None:
            self.lbs = [sr.name for sr in self.ts]
        else:
            self.lbs = lbs

        if plotargs is None:
            self.plotargs = [{} for sr in self.ts]
        else:
            self.plotargs = plotargs

        self._create_axes()

        if plot==True:
            self.plotheads()


    def __repr__(self):
        return "Plot multiple groundwater head time series"

    
    def get_fig(self):
        """Return figure """
        return self._fig

    def mindate(self):
        """Return very first date in list of heads series"""
        return np.array([(sr.index.date).min() for sr in self.ts]).min()


    def maxdate(self):
        """Return very last date in list of heads series"""
        return np.array([(sr.index.date).max() for sr in self.ts]).max()


    def nyears(self):
        """Return maximum number of years with measurements """
        return np.array([len(set(sr.index.year)) for sr in self.ts]).max()


    def _validate_series(self):
        """Replace all NaN values with zero"""
        for i,sr in enumerate(self.ts):
            if len(sr[sr.notnull()])==0:
                self.ts[i] = sr.fillna(0)
                # TODO: user warning


    def _create_axes(self):
        """Create figuer object with axes"""

        #self.fig = plt.figure(figsize=(9, 8), dpi= 80, facecolor='#eeefff', 
        #                        edgecolor='k')
        self._fig = plt.figure(figsize=(8, 5),facecolor='#eeefff')

        if self.mps is None: ## and len(description)==0:

            self._axmp = None
            self._axgws = self._fig.add_axes([0.1,0.1,0.8,0.9])
            self.axeslist = {"axgws":self._axgws}

        else:

            # add graph with reference and level height 
            #  [left, bottom, width, height] values in 0-1 relative figure coordinates:
            self._axmp  = self._fig.add_axes([0.1,0.77,0.8,0.15])
            self._axgws = self._fig.add_axes([0.1,0.1,0.8,0.65])

            self.axeslist = {"axmp":self._axmp,"axgws":self._axgws}


    def plotheads(self,title=None,xlabel=None,ylabel=None,xlim=None,
        ylim=None,colors=None):
        """Plot groundwater heads and reference graph"""

        if title is not None:
            self.title = title

        if xlabel is not None:
            self.xlabel = xlabel

        if ylabel is not None:
            self.ylabel = ylabel

        if xlim is not None:
            self.xlim = xlim
            ##self._set_xaxlimits(xlim)
            self._set_xlim(xlim)

        if ylim is not None:
            self.ylim = ylim
            #self._set_yax_limits()

        if colors is not None:
            self.colors = colors

        ##
        #ts.reverse()
        #lbs.reverse()

        # plot groundwater heads in lower plot
        for i,sr in enumerate(self.ts):

            if not 'color' in self.plotargs[i].keys(): 
                self.plotargs[i]['color']=self.clr[i]
            self.plotargs[i]['lw']=0.9

            #self._axgws = sr.dropna().plot(label=self.lbs[i], 
            #                      **self.plotargs[i])
            x = sr.dropna().index.values
            y = sr.dropna().values
            self._axgws.plot(x, y, label=self.lbs[i], **self.plotargs[i])

        # adjust plot appearance
        self._set_xax_limits(xlim=self.xlim)
        self._set_xax_locators()
        self._set_yax_limits()
        self._set_ticklabels()
        self._set_axlabels()
        self._set_legend()
        self._plot_annotations()

        # plot reference change in upper plot
        if self.mps is not None:
            self._reference_graph()

        return self._axgws


    def _set_yax_limits(self):
        """ Set yax limits for groundwater level """

        if self.ylim is None:

            ymin = min([min(sr.dropna()) for sr in self.ts])
            ymax = max([max(sr.dropna()) for sr in self.ts])

            if ymin==0 and ymax==0: #all series are empty
                ymin = -1.0
                ymax = 1.0

            #ymin = (floor(ymin/10.))*10. # round to decimals
            #ymax = (ceil(ymax/10.))*10.  # round to decimals

            self.ylim = [ymin,ymax]

        self._axgws.set_ylim(self.ylim)

        if self.ref=="surface":
            self._axgws.set_ylim(self.ylim[::-1])


    def _set_xlim(self,xlim=None):
        """Set xlim (mindate,maxdate)"""

        if xlim is None:
            xlim = [self.mindate(),self.maxdate()]

        if not isinstance(xlim,list):
            warnings.warn(f'xlim must be list [mindate,maxdate] '
                f'not {type(xlim)} ')
            xlim = [self.mindate(),self.maxdate()]

        if xlim[0] is None:
            xlim[0] = self.mindate()

        if xlim[1] is None:
            xlim[1] = self.maxdate()

        if isinstance(xlim[0],str):
            xlim[0] = dt.datetime.strptime(xlim[0],'%d-%m-%Y')

        if isinstance(xlim[1],str):
            xlim[1] = dt.datetime.strptime(xlim[0],'%d-%m-%Y')

        if isinstance(xlim[0],int):
            xlim[0] = dt.datetime(year=int(xlim[0]), month=1, day=1)

        if isinstance(xlim[1],int):
            xlim[1] = dt.datetime(year=int(xlim[1]), month=1, day=1)

        """
        if type(self.xlim[0])==str and type(self.xlim[1])==str:
            self.xlim[0] = dt.datetime.strptime(self.xlim[0],'%d-%m-%Y')
            self.xlim[1] = dt.datetime.strptime(self.xlim[1],'%d-%m-%Y')

        if type(self.xlim[0])==int and type(self.xlim[1])==int:
            self.xlim[0] = dt.datetime(year=int(self.xlim[0]), 
                           month=1, day=1)
            self.xlim[1] = dt.datetime(year=int(self.xlim[1]), 
                           month=12, day=31)
        """
        self.xlim = xlim

    def _set_xax_limits(self,xlim=None):
        """ Set xax limits for dates) """

        self._set_xlim(xlim)

        for ax in self.axeslist:
            self.axeslist[ax].set_xlim(self.xlim)


    def _set_xax_locators(self):

        # set xaxis date locators
        self.xaxyears = self.xlim[1].year-self.xlim[0].year+1
        for ax in self.axeslist:
            self.axeslist[ax].set_xlim(self.xlim)

            #years = mdates.YearLocator()   # every year
            #months = mdates.MonthLocator()  # every month
            #years_fmt = mdates.DateFormatter('%Y')
            #self.axeslist[ax].xaxis.set_major_formatter(years_fmt)

            if self.xaxyears in range(0,3):
 
                self.axeslist[ax].xaxis.set_major_locator(
                    YearLocator(1, month=1,day=1))
                self.axeslist[ax].xaxis.set_minor_locator(
                    MonthLocator(bymonth=[1,4,7,10]))

                self.axeslist[ax].xaxis.set_major_formatter(
                    DateFormatter('%Y-%b'))
                self.axeslist[ax].xaxis.set_minor_formatter(
                    DateFormatter('%b'))

            
            if self.xaxyears in range(3,5):

                self.axeslist[ax].xaxis.set_major_locator(
                    YearLocator(1, month=1,day=1))
                self.axeslist[ax].xaxis.set_minor_locator(
                     MonthLocator(bymonth=[4,7,10]))

                self.axeslist[ax].xaxis.set_major_formatter(
                    DateFormatter('%Y'))
                self.axeslist[ax].xaxis.set_minor_formatter(
                    DateFormatter('%b'))
               
            if self.xaxyears in range(5,11):

                self.axeslist[ax].xaxis.set_major_locator(
                     YearLocator(1, month=1,day=1))
                self.axeslist[ax].xaxis.set_minor_locator(
                     YearLocator(1, month=1,day=1))

                self.axeslist[ax].xaxis.set_major_formatter(
                    DateFormatter('%Y'))
                self.axeslist[ax].xaxis.set_minor_formatter(
                    NullFormatter())

            if self.xaxyears in range(11,25):
                self.axeslist[ax].xaxis.set_major_locator(
                     YearLocator(5, month=1,day=1))
                self.axeslist[ax].xaxis.set_minor_locator(
                     YearLocator(1, month=1,day=1))

                self.axeslist[ax].xaxis.set_major_formatter(
                    DateFormatter('%Y'))
                self.axeslist[ax].xaxis.set_minor_formatter(
                    NullFormatter())


            if self.xaxyears in range(25,50):

                self.axeslist[ax].xaxis.set_major_locator(
                     YearLocator(5, month=1,day=1))

                self.axeslist[ax].xaxis.set_major_formatter(
                    DateFormatter('%Y'))

            if self.xaxyears in range(50,1000):

                self.axeslist[ax].xaxis.set_major_locator(
                     YearLocator(10, month=1,day=1))
                     
                self.axeslist[ax].xaxis.set_major_formatter(
                    DateFormatter('%Y'))



    def _set_ticklabels(self):
    
        for label in self._axgws.get_xticklabels():
            label.set_color('k')
            label.set_fontsize = 5.0
            label.set_visible(True)
            label.set_rotation(0)
            label.set_horizontalalignment('center')

        for label in self._axgws.get_yticklabels():
            label.set_color('k')
            label.set_fontsize = 5.0
            label.set_visible(True)


    def _set_axlabels(self):

        if self.ylabel:
            plt.ylabel(self.ylabel, size = 11.0)
        else:
            plt.ylabel('grondwaterstand',size = 11.0)

        self.axeslist['axgws'].set_xlabel('')


    def _set_legend(self):

        # define legend labels
        self._lbs = [series.name for series in self.ts]
        if len(self._lbs)==0:
            return None

        ##plt.legend(shadow = False, loc = 4)

        # calculate position of legend
        if len(self._lbs)<=4:
            ncol=4
            bbox = [0., -0.20, 1., .15]
        elif len(self._lbs)<=8:
            ncol=4
            bbox = [0., -0.25, 1., .15]

        # explanation bbox:
        # plt.legend(bbox_to_anchor=(x0, y0, width, height), loc=4
        # width and height are the width and the height of the 
        # legend box, and (x0, y0) is the coordinate of the loc of 
        # the bounding box. loc=4 is lower right

        self.axeslist['axgws'].legend(loc='lower center', 
             ncol=ncol, bbox_to_anchor=(bbox), 
             mode="expand", borderaxespad=0.,frameon=False)

        # plot legend texts
        ##ltext = plt.gca().get_legend().get_texts()
        ltext = self.axeslist['axgws'].get_legend().get_texts()
        for i in range(len(ltext)):
            plt.setp(ltext[i], fontsize = 8.0, color = 'k')

        # set the linewidth of each legend object
        """
        leg = self.axeslist['axgws'].get_legend()
        for legobj in leg.legendHandles:
            legobj.set_linewidth(2.0)
        """

        leg = self.axeslist['axgws'].get_legend()
        for line in leg.get_lines():
            line.set_linewidth(2.5)


    def _reference_graph(self):
        """ Plot timeseries with relative changes """

        # plot reference line on top graph
        #self.mps.plot(ax=self._axmp, color=self.clr[1],
        #              lw=1.5)

        x = self.mps.index.values
        y = self.mps.values
        self._axmp.plot(x, y, color = self.clr[1], lw=1.5)

        # set x-axis equal to grondwwater series
        self._axmp.set_xlim(self.xlim)

        # set grid on
        self._axmp.grid(True,which="both",ls="-")

        # x-ticklabels for mp graph invisible
        for label in self._axmp.get_xticklabels():
            label.set_visible(False)

        # x-axis title for reference graph invisible
        self._axmp.set_xlabel('')


    def _plot_annotations(self):
        """ Plot figure title and measurment periode """

        self._axgws.grid(True,which="major",ls="-")
        self._axgws.grid(True,which="minor",ls=":")

        if self.title and (self.mps is None):
            plt.text(0.0,1.02,self.title,transform=self._axgws.transAxes)
        elif self.title:
            plt.text(0.0,1.1,self.title,transform=self._axmp.transAxes)

        # plot datespan right of graph 90 degrees upward
        #timespan = self.mindate().strftime("%d-%m-%Y")+" t/m " \
        #           +self.maxdate().strftime("%d-%m-%Y")
        timespan = self.xlim[0].strftime("%d-%m-%Y")+" t/m " \
                   +self.xlim[1].strftime("%d-%m-%Y")

        self.axeslist['axgws'].text(1.03, 0., timespan, 
             horizontalalignment='center',
             verticalalignment='bottom',rotation='vertical',
             transform=self.axeslist['axgws'].transAxes)


    def save(self,filename,dpi=None):
        """Save figure to file

        Parameters
        ----------
        filename : str
            filename for saving figure (including extension)

        dpi : number, default 200.0
            dpi for output figure

        """

        if dpi is None:
            dpi = 200.0 # default dpi is 100.0
        self._fig.set_dpi(dpi)
        self._fig.set_size_inches(160*self.mm,80*self.mm)
        self._fig.savefig(
                filename,dpi=dpi, 
                facecolor='w', 
                edgecolor='w', 
                bbox_inches="tight") #additional_artists=self.addart,
